import math
from collections import OrderedDict
from copy import deepcopy
from types import SimpleNamespace
from typing import Any, List, Tuple, Union
from warnings import warn

try:
    from typing import Self
except ImportError:
    from typing import TypeVar

    Self = TypeVar('Self', bound='Sequence')

import numpy as np
from scipy.interpolate import PPoly

from pypulseq import __version__, eps
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.check_timing import check_timing as ext_check_timing
from pypulseq.check_timing import print_error_report
from pypulseq.decompress_shape import decompress_shape
from pypulseq.event_lib import EventLibrary
from pypulseq.opts import Opts
from pypulseq.Sequence import block
from pypulseq.Sequence.calc_grad_spectrum import calculate_gradient_spectrum
from pypulseq.Sequence.calc_pns import calc_pns
from pypulseq.Sequence.ext_test_report import ext_test_report
from pypulseq.Sequence.install import detect_scanner
from pypulseq.Sequence.read_seq import read
from pypulseq.Sequence.write_seq import write as write_seq
from pypulseq.Sequence.write_seq import write_v141 as write_seq_v141
from pypulseq.utils.cumsum import cumsum
from pypulseq.utils.paper_plot import paper_plot as ext_paper_plot
from pypulseq.utils.seq_plot import SeqPlot
from pypulseq.utils.tracing import format_trace, trace, trace_enabled

major, minor, revision = __version__.split('.')[:3]


class Sequence:
    """
    Generate sequences and read/write sequence files. This class defines properties and methods to define a complete MR
    sequence including RF pulses, gradients, ADC events, etc. The class provides an implementation of the open MR
    sequence format defined by the Pulseq project. See http://pulseq.github.io/.

    See also `demo_read.py`, `demo_write.py`.
    """

    version_major = int(major)
    version_minor = int(minor)
    version_revision = revision

    def __init__(self, system: Union[Opts, None] = None, use_block_cache: bool = True):
        if system is None:
            system = Opts()

        # =========
        # EVENT LIBRARIES
        # =========
        self.adc_library = EventLibrary()
        self.delay_library = EventLibrary()
        self.extensions_library = EventLibrary()
        self.grad_library = EventLibrary()
        self.label_inc_library = EventLibrary()
        self.label_set_library = EventLibrary()
        self.rf_library = EventLibrary()
        self.shape_library = EventLibrary(numpy_data=True)
        self.trigger_library = EventLibrary()
        self.soft_delay_library = EventLibrary()

        # =========
        # OTHER
        # =========
        self.system = system

        self.block_events = OrderedDict()
        self.block_trace = OrderedDict()
        self.use_block_cache = use_block_cache
        self.block_cache = {}
        self.next_free_block_ID = 1
        self.definitions = {}

        self.rf_raster_time = self.system.rf_raster_time
        self.grad_raster_time = self.system.grad_raster_time
        self.adc_raster_time = self.system.adc_raster_time
        self.block_duration_raster = self.system.block_duration_raster
        self.set_definition('AdcRasterTime', self.adc_raster_time)
        self.set_definition('BlockDurationRaster', self.block_duration_raster)
        self.set_definition('GradientRasterTime', self.grad_raster_time)
        self.set_definition('RadiofrequencyRasterTime', self.rf_raster_time)
        self.signature_type = ''
        self.signature_file = ''
        self.signature_value = ''
        self.rf_id_to_name_map = {}
        self.adc_id_to_name_map = {}
        self.grad_id_to_name_map = {}

        self.block_durations = {}
        self.extension_numeric_idx = []
        self.extension_string_idx = []
        self.soft_delay_hints = {}

    def __str__(self) -> str:
        s = 'Sequence:'
        s += '\nshape_library: ' + str(self.shape_library)
        s += '\nrf_library: ' + str(self.rf_library)
        s += '\ngrad_library: ' + str(self.grad_library)
        s += '\nadc_library: ' + str(self.adc_library)
        s += '\ndelay_library: ' + str(self.delay_library)
        s += '\nextensions_library: ' + str(self.extensions_library)
        s += '\nrf_raster_time: ' + str(self.rf_raster_time)
        s += '\ngrad_raster_time: ' + str(self.grad_raster_time)
        s += '\nblock_events: ' + str(len(self.block_events))
        return s

    def adc_times(self, time_range: Union[List[float], None] = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Return time points of ADC sampling points.

        Returns
        -------
        t_adc: np.ndarray
            Contains times of all ADC sample points.
        fp_adc : np.ndarray
            Contains frequency and phase offsets of each ADC object (not samples).
        """
        # Collect ADC timing data
        t_adc = []
        fp_adc = []

        curr_dur = 0
        if time_range is None:
            blocks = self.block_events
        else:
            if len(time_range) != 2:
                raise ValueError('Time range must be list of two elements')
            if time_range[0] > time_range[1]:
                raise ValueError('End time of time_range must be after begin time')

            # Calculate end times of each block
            bd = np.array(list(self.block_durations.values()))
            t = np.cumsum(bd)
            # Search block end times for start of time range
            begin_block = np.searchsorted(t, time_range[0])
            # Search block begin times for end of time range
            end_block = np.searchsorted(t - bd, time_range[1], side='right')
            blocks = list(self.block_durations.keys())[begin_block:end_block]
            curr_dur = t[begin_block] - bd[begin_block]

        for block_counter in blocks:
            block = self.get_block(block_counter)

            if block.adc is not None:
                t_adc.append((np.arange(block.adc.num_samples) + 0.5) * block.adc.dwell + block.adc.delay + curr_dur)
                fp_adc.append([block.adc.freq_offset, block.adc.phase_offset])

            curr_dur += self.block_durations[block_counter]

        if t_adc == []:
            # If there are no ADCs, make sure the output is the right shape
            t_adc = np.zeros(0)
            fp_adc = np.zeros((0, 2))
        else:
            t_adc = np.concatenate(t_adc)
            fp_adc = np.array(fp_adc)

        return t_adc, fp_adc

    def add_block(self, *args: SimpleNamespace) -> None:
        """
        Add a new block/multiple events to the sequence.

        Parameters
        ----------
        *args : SimpleNamespace
            Event objects to be added as a block to the sequence. For delays,
            use `pypulseq.make_delay()` instead of raw float values.

        See Also
        --------
        pypulseq.make_delay : Create delay events
        pypulseq.make_adc : Create ADC events
        pypulseq.make_trapezoid : Create trapezoid gradient events
        pypulseq.make_sinc_pulse : Create sinc RF pulse events
        pypulseq.make_soft_delay : Create soft delay events
        """
        # Validate that no raw floats are passed directly by users
        for arg in args:
            if isinstance(arg, float):
                raise ValueError(
                    f'Raw float values are not allowed in add_block(). Use pp.make_delay({arg}) for delays.'
                )

        if trace_enabled():
            self.block_trace[self.next_free_block_ID] = SimpleNamespace(block=trace())

        block.set_block(self, self.next_free_block_ID, *args)
        self.next_free_block_ID += 1

    def calculate_gradient_spectrum(
        self,
        max_frequency: float = 2000.0,
        window_width: float = 0.05,
        frequency_oversampling: float = 3.0,
        time_range: Union[List[float], None] = None,
        plot: bool = True,
        combine_mode: str = 'max',
        use_derivative: bool = False,
        acoustic_resonances: Union[List[dict], None] = None,
    ) -> Tuple[List[np.ndarray], np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculates the gradient spectrum of the sequence. Returns a spectrogram
        for each gradient channel, as well as a root-sum-squares combined
        spectrogram.

        Works by splitting the sequence into windows that are 'window_width'
        long and calculating the fourier transform of each window. Windows
        overlap 50% with the previous and next window. When 'combine_mode' is
        not 'none', all windows are combined into one spectrogram.

        Parameters
        ----------
        max_frequency : float, optional
            Maximum frequency to include in spectrograms. The default is 2000.
        window_width : float, optional
            Window width (in seconds). The default is 0.05.
        frequency_oversampling : float, optional
            Oversampling in the frequency dimension, higher values make
            smoother spectrograms. The default is 3.
        time_range : List[float], optional
            Time range over which to calculate the spectrograms as a list of
            two timepoints (in seconds) (e.g. [1, 1.5])
            The default is None.
        plot : bool, optional
            Whether to plot the spectograms. The default is True.
        combine_mode : str, optional
            How to combine all windows into one spectrogram, options:
                'max', 'mean', 'sos' (root-sum-of-squares), 'none' (no combination)
            The default is 'max'.
        use_derivative : bool, optional
            Whether the use the derivative of the gradient waveforms instead of the
            gradient waveforms for the gradient spectrum calculations. The default
            is False
        acoustic_resonances : List[dict], optional
            Acoustic resonances as a list of dictionaries with 'frequency' and
            'bandwidth' elements. Only used when plot==True. The default is [].

        Returns
        -------
        spectrograms : List[np.ndarray]
            List of spectrograms per gradient channel.
        spectrogram_sos : np.ndarray
            Root-sum-of-squares combined spectrogram over all gradient channels.
        frequencies : np.ndarray
            Frequency axis of the spectrograms.
        times : np.ndarray
            Time axis of the spectrograms (only relevant when combine_mode == 'none').

        """
        if acoustic_resonances is None:
            acoustic_resonances = []

        return calculate_gradient_spectrum(
            self,
            max_frequency=max_frequency,
            window_width=window_width,
            frequency_oversampling=frequency_oversampling,
            time_range=time_range,
            plot=plot,
            combine_mode=combine_mode,
            use_derivative=use_derivative,
            acoustic_resonances=acoustic_resonances,
        )

    def calculate_kspace(
        self,
        trajectory_delay: Union[float, List[float], np.ndarray] = 0.0,
        gradient_offset: Union[float, List[float], np.ndarray] = 0.0,
    ) -> Tuple[np.ndarray, np.ndarray, List[float], List[float], np.ndarray]:
        """
        Calculates the k-space trajectory of the entire pulse sequence.

        Parameters
        ----------
        trajectory_delay : float or list or numpy.ndarray, default=0
            Compensation factor in seconds (s) to align ADC and gradients in the reconstruction.
            If trajectory_delay is a single value, this value will be used for all gradient channels.
            If trajectory_delay is a list or array, it is expected to have the same length as the number of gradient
            channels and the first element is applied to the first gradient channel, the second to the second, and so on.
        gradient_offset : float or list or numpy.ndarray, default=0
            Simulates background gradients (specified in Hz/m)
            If gradient_offset is a single value, this value will be used for all gradient channels.
            If gradient_offset is a list or array, it is expected to have the same length as the number of gradient
            channels and the first element is applied to the first gradient channel, the second to the second, and so on.

        Returns
        -------
        k_traj_adc : numpy.array
            K-space trajectory sampled at `t_adc` timepoints.
        k_traj : numpy.array
            K-space trajectory of the entire pulse sequence.
        t_excitation : List[float]
            Excitation timepoints.
        t_refocusing : List[float]
            Refocusing timepoints.
        t_adc : numpy.array
            Sampling timepoints.
        """
        if np.any(np.abs(trajectory_delay) > 100e-6):
            raise Warning(f'Trajectory delay of {trajectory_delay * 1e6} us is suspiciously high')

        total_duration = sum(self.block_durations.values())

        t_excitation, fp_excitation, t_refocusing, _ = self.rf_times()
        t_adc, _ = self.adc_times()

        # Convert data to piecewise polynomials
        gw_pp = self.get_gradients(trajectory_delay, gradient_offset)
        ng = len(gw_pp)

        # Calculate slice positions.
        # For now we entirely rely on the excitation -- ignoring complicated interleaved refocused sequences
        if len(t_excitation) > 0:
            # Position in x, y, z
            slice_pos = np.zeros((ng, len(t_excitation)))
            for j in range(ng):
                if gw_pp[j] is None:
                    slice_pos[j] = np.nan
                else:
                    # Check for divisions by zero to avoid numpy warning
                    divisor = np.array(gw_pp[j](t_excitation))
                    slice_pos[j, divisor != 0.0] = fp_excitation[0, divisor != 0.0] / divisor[divisor != 0.0]
                    slice_pos[j, divisor == 0.0] = np.nan

            slice_pos[~np.isfinite(slice_pos)] = 0  # Reset undefined to 0
        else:
            slice_pos = []

        # Integrate waveforms as PPs to produce gradient moments
        gm_pp = []
        tc = []
        for i in range(ng):
            if gw_pp[i] is None:
                gm_pp.append(None)
                continue

            gm_pp.append(gw_pp[i].antiderivative())
            tc.append(gm_pp[i].x)
            # "Sample" ramps for display purposes.  Otherwise piecewise-linear display (plot) fails
            ii = np.flatnonzero(np.abs(gm_pp[i].c[0, :]) > 1e-7 * self.system.max_slew)

            # Do nothing if there are no ramps
            if ii.shape[0] == 0:
                continue

            starts = np.int64(np.floor((gm_pp[i].x[ii] + eps) / self.grad_raster_time))
            ends = np.int64(np.ceil((gm_pp[i].x[ii + 1] - eps) / self.grad_raster_time))

            # Create all ranges starts[0]:ends[0], starts[1]:ends[1], etc.
            lengths = ends - starts + 1
            inds = np.ones((lengths).sum())
            # Calculate output index where each range will start
            start_inds = np.cumsum(np.concatenate(([0], lengths[:-1])))
            # Create element-wise differences that will cumsum into
            # the final indices: [starts[0], 1, 1, starts[1]-starts[0]-lengths[0]+1, 1, etc.]
            inds[start_inds] = np.concatenate(([starts[0]], np.diff(starts) - lengths[:-1] + 1))

            tc.append(np.cumsum(inds) * self.grad_raster_time)
        if tc != []:
            tc = np.concatenate(tc)

        t_acc = 1e-10  # Temporal accuracy
        t_acc_inv = 1 / t_acc
        # tc = self.__flatten_jagged_arr(tc)
        t_ktraj = t_acc * np.unique(
            np.round(
                t_acc_inv
                * np.array(
                    [
                        *tc,
                        0,
                        *np.asarray(t_excitation) - 2 * self.rf_raster_time,
                        *np.asarray(t_excitation) - self.rf_raster_time,
                        *t_excitation,
                        *np.asarray(t_refocusing) - self.rf_raster_time,
                        *t_refocusing,
                        *t_adc,
                        total_duration,
                    ]
                )
            )
        )

        i_excitation = np.searchsorted(t_ktraj, t_acc * np.round(t_acc_inv * np.asarray(t_excitation)))
        i_refocusing = np.searchsorted(t_ktraj, t_acc * np.round(t_acc_inv * np.asarray(t_refocusing)))
        i_adc = np.searchsorted(t_ktraj, t_acc * np.round(t_acc_inv * np.asarray(t_adc)))

        i_periods = np.unique([0, *i_excitation, *i_refocusing, len(t_ktraj) - 1])
        if len(i_excitation) > 0:
            ii_next_excitation = 0
        else:
            ii_next_excitation = -1
        if len(i_refocusing) > 0:
            ii_next_refocusing = 0
        else:
            ii_next_refocusing = -1

        k_traj = np.zeros((ng, len(t_ktraj)))
        for i in range(ng):
            if gw_pp[i] is None:
                continue

            it = np.where(
                np.logical_and(
                    t_ktraj >= t_acc * round(t_acc_inv * gm_pp[i].x[0]),
                    t_ktraj <= t_acc * round(t_acc_inv * gm_pp[i].x[-1]),
                )
            )[0]
            k_traj[i, it] = gm_pp[i](t_ktraj[it])
            if t_ktraj[it[-1]] < t_ktraj[-1]:
                k_traj[i, it[-1] + 1 :] = k_traj[i, it[-1]]

        # Convert gradient moments to k-space positions
        dk = -k_traj[:, 0]
        for i in range(len(i_periods) - 1):
            i_period = i_periods[i]
            i_period_end = i_periods[i + 1]
            if ii_next_excitation >= 0 and i_excitation[ii_next_excitation] == i_period:
                if abs(t_ktraj[i_period] - t_excitation[ii_next_excitation]) > t_acc:
                    raise Warning(
                        f'abs(t_ktraj[i_period]-t_excitation[ii_next_excitation]) < {t_acc} failed for ii_next_excitation={ii_next_excitation} error={t_ktraj(i_period) - t_excitation(ii_next_excitation)}'
                    )
                dk = -k_traj[:, i_period]
                if i_period > 0:
                    # Use nans to mark the excitation points since they interrupt the plots
                    k_traj[:, i_period - 1] = np.nan
                # -1 on len(i_excitation) for 0-based indexing
                ii_next_excitation = min(len(i_excitation) - 1, ii_next_excitation + 1)
            elif ii_next_refocusing >= 0 and i_refocusing[ii_next_refocusing] == i_period:
                # dk = -k_traj[:, i_period]
                dk = -2 * k_traj[:, i_period] - dk
                # -1 on len(i_excitation) for 0-based indexing
                ii_next_refocusing = min(len(i_refocusing) - 1, ii_next_refocusing + 1)

            k_traj[:, i_period:i_period_end] = k_traj[:, i_period:i_period_end] + dk[:, None]

        k_traj[:, i_period_end] = k_traj[:, i_period_end] + dk
        k_traj_adc = k_traj[:, i_adc]

        return k_traj_adc, k_traj, t_excitation, t_refocusing, t_adc

    def calculate_kspacePP(
        self,
        trajectory_delay: Union[float, List[float], np.ndarray] = 0,
        gradient_offset: Union[float, List[float], np.ndarray] = 0,
    ) -> Tuple[np.array, np.array, np.array, np.array, np.array]:
        warn(
            'Sequence.calculate_kspacePP has been deprecated, use calculate_kspace instead',
            DeprecationWarning,
            stacklevel=2,
        )
        return self.calculate_kspace(trajectory_delay, gradient_offset)

    def calculate_pns(
        self,
        hardware: SimpleNamespace,
        time_range: Union[List[float], None] = None,
        do_plots: bool = True,
    ) -> Tuple[bool, np.ndarray, np.ndarray, np.ndarray]:
        """
        Calculate PNS using safe model implementation by Szczepankiewicz and Witzel
        See http://github.com/filip-szczepankiewicz/safe_pns_prediction

        Returns pns levels due to respective axes (normalized to 1 and not to 100#)

        Parameters
        ----------
        hardware : SimpleNamespace
            Hardware specifications. See safe_example_hw() from
            the safe_pns_prediction package. Alternatively a text file
            in the .asc format (Siemens) can be passed, e.g. for Prisma
            it is MP_GPA_K2309_2250V_951A_AS82.asc (we leave it as an
            exercise to the interested user to find were these files
            can be acquired from)
        do_plots : bool, optional
            Plot the results from the PNS calculations. The default is True.

        Returns
        -------
        ok : bool
            Boolean flag indicating whether peak PNS is within acceptable limits
        pns_norm : numpy.array [N]
            PNS norm over all gradient channels, normalized to 1
        pns_components : numpy.array [Nx3]
            PNS levels per gradient channel
        t_pns : np.array [N]
            Time axis for the pns_norm and pns_components arrays
        """
        return calc_pns(self, hardware, time_range=time_range, do_plots=do_plots)

    def check_timing(self, print_errors=False) -> Tuple[bool, List[SimpleNamespace]]:
        """
        Checks timing of all blocks and objects in the sequence optionally returns the detailed error log.

        Returns
        -------
        is_ok : bool
            Boolean flag indicating timing errors.
        error_report : List[SimpleNamespace]
            Error report in case of timing errors.
        """
        is_ok, error_report = ext_check_timing(self)

        if not is_ok and print_errors:
            print_error_report(self, error_report)

        return is_ok, error_report

    def duration(self) -> Tuple[int, int, np.ndarray]:
        """
        Returns the total duration of this sequence, and the total count of blocks and events.

        Returns
        -------
        duration : int
            Duration of this sequence in seconds (s).
        num_blocks : int
            Number of blocks in this sequence.
        event_count : np.ndarray
            Number of events in this sequence.
        """
        num_blocks = len(self.block_events)
        event_count = np.zeros(len(next(iter(self.block_events.values()))))
        duration = 0
        for block_counter in self.block_events:
            event_count += self.block_events[block_counter] > 0
            duration += self.block_durations[block_counter]

        return duration, num_blocks, event_count

    def evaluate_labels(self, init: Union[dict, None] = None, evolution: str = 'none') -> dict:
        """
        Evaluate label values of the entire sequence.

        When no evolution is given, returns the label values at the end of the
        sequence. Returns a dictionary with keys named after the labels used
        in the sequence. Only the keys corresponding to the labels actually
        used are created.
        E.g. labels['LIN'] == 4

        When evolution is given, labels are tracked through the sequence. See
        below for options for different types of evolutions. The resulting
        dictionary will contain arrays of the label values.
        E.g. labels['LIN'] == np.array([0,1,2,3,4])

        Initial values for the labels can be given with the 'init' parameter.
        Useful if evaluating labels block-by-block.

        Parameters
        ----------
        init : dict, optional
            Dictionary containing initial label values. The default is None.
        evolution : str, optional
            Flag to specify tracking of label evolutions.
            Must be one of: 'none', 'adc', 'label', 'blocks' (default = 'none')
            'blocks': Return label values for all blocks.
            'adc':    Return label values only for blocks containing ADC events.
            'label':  Return label values only for blocks where labels are
                      manipulated.

        Returns
        -------
        labels : dict
            Dictionary containing label values.
            If evolution == 'none', the dictionary values only contains the
            final label value.
            Otherwise, the dictionary values are arrays of label evolutions.
            Only the labels that are used in the sequence are created in the
            dictionary.

        """
        labels = init or {}
        label_evolution = []

        # TODO: MATLAB implementation includes block_range parameter. But in
        #       general we cannot assume linear block ordering. Could include
        #       time_range like in other sequence functions. Or a blocks
        #       parameter to specify which blocks to loop over?
        for block_counter in self.block_events:
            block = self.get_block(block_counter)

            if block.label is not None:
                # Current block has labels
                for lab in block.label.values():
                    if lab.type == 'labelinc':
                        # Increment label
                        if lab.label not in labels:
                            labels[lab.label] = 0

                        labels[lab.label] += lab.value
                    else:
                        # Set label
                        labels[lab.label] = lab.value

                if evolution == 'label':
                    label_evolution.append(dict(labels))

            if evolution == 'blocks' or (evolution == 'adc' and block.adc is not None):
                label_evolution.append(dict(labels))

        # Convert evolutions into label dictionary
        if len(label_evolution) > 0:
            for lab in labels:
                labels[lab] = np.array([e.get(lab, 0) for e in label_evolution])

        return labels

    import numpy as np

    def find_block_by_time(self, t: float) -> int:
        """
        Find the index of the block containing time `t`.

        Parameters
        ----------
        t : float
            Time (in seconds) to locate within the sequence.

        Returns
        -------
        int or None
            Index of the block that contains the given time, or None if out of range.
        """
        cumsum_durations = np.cumsum(list(self.block_durations.values()))
        block_index = np.searchsorted(cumsum_durations, t, side='right').item()

        if block_index >= len(self.block_durations):
            return None

        if self.block_durations[block_index] <= 0:
            raise ValueError('Block duration cannot be negative')

        return block_index

    def flip_grad_axis(self, axis: str) -> None:
        """
        Invert all gradients along the corresponding axis/channel. The function acts on all gradient objects already
        added to the sequence object.

        Parameters
        ----------
        axis : str
            Gradients to invert or scale. Must be one of 'x', 'y' or 'z'.
        """
        self.mod_grad_axis(axis, modifier=-1)

    def get_raw_block_content_IDs(self, block_index: int) -> SimpleNamespace:
        """
        Returns PyPulseq block content IDs at `block_index` position in `self.block_events`.

        No block events are created, only the IDs of the objects are returned.

        See Also
        --------
        - `pypulseq.Sequence.sequence.Sequence.get_block()`.

        Parameters
        ----------
        block_index : int
            Index of block to be retrieved from `Sequence`.

        Returns
        -------
        SimpleNamespace
            PyPulseq block content IDs at 'block_index' position in `self.block_events`.
        """
        return block.get_raw_block_content_IDs(self, block_index)

    def get_block(self, block_index: int) -> SimpleNamespace:
        """
        Return a block of the sequence  specified by the index. The block is created from the sequence data with all
        events and shapes decompressed.

        See Also
        --------
        - `pypulseq.Sequence.sequence.Sequence.set_block()`.
        - `pypulseq.Sequence.sequence.Sequence.add_block()`.

        Parameters
        ----------
        block_index : int
            Index of block to be retrieved from `Sequence`.

        Returns
        -------
        SimpleNamespace
            Event identified by `block_index`.
        """
        return block.get_block(self, block_index)

    def get_definition(self, key: str) -> str:
        """
        Return value of the definition specified by the key. These definitions can be added manually or read from the
        header of a sequence file defined in the sequence header. An empty array is returned if the key is not defined.

        See also `pypulseq.Sequence.sequence.Sequence.set_definition()`.

        Parameters
        ----------
        key : str
            Key of definition to retrieve.

        Returns
        -------
        str
            Definition identified by `key` if found, else returns ''.
        """
        if key in self.definitions:
            return self.definitions[key]
        else:
            return ''

    def get_extension_type_ID(self, extension_string: str) -> int:
        """
        Get numeric extension ID for `extension_string`. Will automatically create a new ID if unknown.

        Parameters
        ----------
        extension_string : str
            Given string extension ID.

        Returns
        -------
        extension_id : int
            Numeric ID for given string extension ID.

        """
        if extension_string not in self.extension_string_idx:
            if len(self.extension_numeric_idx) == 0:
                extension_id = 1
            else:
                extension_id = 1 + max(self.extension_numeric_idx)

            self.extension_numeric_idx.append(extension_id)
            self.extension_string_idx.append(extension_string)
            assert len(self.extension_numeric_idx) == len(self.extension_string_idx)
        else:
            num = self.extension_string_idx.index(extension_string)
            extension_id = self.extension_numeric_idx[num]

        return extension_id

    def get_extension_type_string(self, extension_id: int) -> str:
        """
        Get string extension ID for `extension_id`.

        Parameters
        ----------
        extension_id : int
            Given numeric extension ID.

        Returns
        -------
        extension_str : str
            String ID for the given numeric extension ID.

        Raises
        ------
        ValueError
            If given numeric extension ID is unknown.
        """
        if extension_id in self.extension_numeric_idx:
            num = self.extension_numeric_idx.index(extension_id)
        else:
            raise ValueError(f'Extension for the given ID - {extension_id} - is unknown.')

        extension_str = self.extension_string_idx[num]
        return extension_str

    def get_gradients(
        self,
        trajectory_delay: Union[float, List[float], np.ndarray] = 0,
        gradient_offset: Union[float, List[float], np.ndarray] = 0,
        time_range: Union[List[float], None] = None,
    ) -> List[PPoly]:
        """
        Get all gradient waveforms of the sequence in a piecewise-polynomial
        format (scipy PPoly). Gradient values can be accessed easily at one or
        more timepoints using `gw_pp[channel](t)` (where t is a float, list of
        floats, or numpy array). Note that PPoly objects return nan for
        timepoints outside the waveform.

        Parameters
        ----------
        trajectory_delay : float or list or numpy.ndarray, default=0
            Compensation factor in seconds (s) to align ADC and gradients in the reconstruction.
            If trajectory_delay is a single value, this value will be used for all gradient channels.
            If trajectory_delay is a list or array, it is expected to have the same length as the number of gradient
            channels and the first element is applied to the first gradient channel, the second to the second, and so on.
        gradient_offset : float or list or numpy.ndarray, default=0
            Simulates background gradients (specified in Hz/m)
            If gradient_offset is a single value, this value will be used for all gradient channels.
            If gradient_offset is a list or array, it is expected to have the same length as the number of gradient
            channels and the first element is applied to the first gradient channel, the second to the second, and so on.

        Returns
        -------
        gw_pp : List[PPoly]
            List of gradient waveforms for each of the gradient channels,
            expressed as scipy PPoly objects.
        """
        if np.any(np.abs(trajectory_delay) > 100e-6):
            raise Warning(f'Trajectory delay of {trajectory_delay * 1e6} us is suspiciously high')

        total_duration = sum(self.block_durations.values())

        gw_data = self.waveforms(time_range=time_range)
        ng = len(gw_data)

        # Gradient delay handling
        if isinstance(trajectory_delay, (int, float)):
            gradient_delays = [trajectory_delay] * ng
        else:
            assert len(trajectory_delay) == ng  # Need to have same number of gradient channels
            gradient_delays = trajectory_delay

        # Gradient offset handling
        if isinstance(gradient_offset, (int, float)):
            gradient_offset = [gradient_offset] * ng
        else:
            assert len(gradient_offset) == ng  # Need to have same number of gradient channels

        # Convert data to piecewise polynomials
        gw_pp = []
        for j in range(ng):
            wave_cnt = gw_data[j].shape[1]
            if wave_cnt == 0:
                if np.abs(gradient_offset[j]) <= eps:
                    gw_pp.append(None)
                    continue
                else:
                    gw = np.array(([0, total_duration], [0, 0]))
            else:
                gw = gw_data[j]

            # Now gw contains the waveform from the current axis
            if np.abs(gradient_delays[j]) > eps:
                gw[0] = gw[0] - gradient_delays[j]  # Anisotropic gradient delay support
            if not np.all(np.isfinite(gw)):
                raise Warning('Not all elements of the generated waveform are finite.')

            teps = 1e-12
            _temp1 = np.array(([gw[0, 0] - 2 * teps, gw[0, 0] - teps], [0, 0]))
            _temp2 = np.array(([gw[0, -1] + teps, gw[0, -1] + 2 * teps], [0, 0]))
            gw = np.hstack((_temp1, gw, _temp2))

            if np.abs(gradient_offset[j]) > eps:
                gw[1, :] += gradient_offset[j]

            gw[1][gw[1] == -0.0] = 0.0

            gw_pp.append(PPoly(np.stack((np.diff(gw[1]) / np.diff(gw[0]), gw[1][:-1])), gw[0], extrapolate=True))
        return gw_pp

    def install(self, target: Union[str, None] = None, clear_cache: bool = False, **kwargs: Any) -> None:
        """Install a sequence to a target scanner.

        The sequence will be installed to a scanner specified by `target`. If `target` is not specified,
        all known scanners will be attempted to be detected. Targets supported by PyPulseq:
            siemens: All siemens targets below
            siemens_nx: Siemens Numaris X
            siemens_n4: Siemens Numaris 4
            siemens_n4_2: Siemens Numeris 4 on IP 192.168.2.2
            siemens_n4_3: Siemens Numeris 4 on IP 192.168.2.3
        Once a scanner is successfully detected, this result will be cached so
        `install` will operate more quickly on successive calls. The cache can
        be cleared by specifying `clear_cache=True`.
        Parameters
        ----------
        target : str, optional
            Target scanner. The default is None.
        clear_cache : bool, optional
            Clear the scanner detection cache. The default is False.
        **kwargs : Any
            Keyword arguments to be passed to the target's `install` function.
        Raises
        ------
        RuntimeError
            If the scanner could not be detected, or if the installation failed.
        """
        name, definition = detect_scanner(target, clear_cache=clear_cache)

        if definition is None:
            raise RuntimeError('Scanner could not be detected')

        if not definition.install(self, **kwargs):
            raise RuntimeError('Sequence install failed')

        print(f'Sequence installed correctly on target `{name}`')

    def mod_grad_axis(self, axis: str, modifier: float) -> None:
        """
        Invert or scale all gradients along the corresponding axis/channel.

        The function acts on all gradient objects already added to the sequence object.
        This modifies the amplitude of gradients while preserving timing parameters
        (rise_time, flat_time, fall_time, delay). The gradient area scales proportionally
        with the amplitude.

        Parameters
        ----------
        axis : str
            Gradients to invert or scale. Must be one of 'x', 'y' or 'z'.
        modifier : float
            Scaling value.

        Raises
        ------
        ValueError
            If invalid `axis` is passed. Must be one of 'x', 'y','z'.
        RuntimeError
            If same gradient event is used on multiple axes.

        Examples
        --------
        Disable phase encoding gradients (y-axis):

        >>> seq.mod_grad_axis('y', 0.0)

        Invert readout gradients (x-axis):

        >>> seq.mod_grad_axis('x', -1.0)

        Reduce slice selection gradients by half (z-axis):

        >>> seq.mod_grad_axis('z', 0.5)

        Double all gradients on x-axis:

        >>> seq.mod_grad_axis('x', 2.0)

        Notes
        -----
        - Only amplitude-related parameters are modified (amplitude, area, first, last)
        - Timing parameters remain unchanged (rise_time, flat_time, fall_time, delay)
        - For arbitrary gradients, the entire waveform is scaled
        - For trapezoid gradients, only the amplitude is scaled
        - Setting modifier to 0.0 effectively disables gradients on that axis
        """
        if axis not in ['x', 'y', 'z']:
            raise ValueError(f"Invalid axis. Must be one of 'x', 'y','z'. Passed: {axis}")

        channel_num = ['x', 'y', 'z'].index(axis)
        other_channels = [0, 1, 2]
        other_channels.remove(channel_num)

        # Go through all event table entries and list gradient objects in the library
        if len(self.block_events) == 0:
            # Empty sequence - nothing to modify
            return

        all_grad_events = np.array(list(self.block_events.values()))
        all_grad_events = all_grad_events[:, 2:5]

        selected_events = np.unique(all_grad_events[:, channel_num])
        selected_events = selected_events[selected_events != 0]
        other_events = np.unique(all_grad_events[:, other_channels])
        if len(np.intersect1d(selected_events, other_events)) > 0:
            raise RuntimeError('mod_grad_axis does not yet support the same gradient event used on multiple axes.')

        for i in range(len(selected_events)):
            event_id = selected_events[i]
            old_data = self.grad_library.data[event_id]
            grad_type = self.grad_library.type[event_id]

            # Convert tuple to list for modification
            grad_data = list(old_data)
            grad_data[0] *= modifier

            if grad_type == 'g' and len(grad_data) == 6:
                # Need to update first and last fields for arbitrary gradients
                # Data structure: (amplitude, first, last, shape_ID1, shape_ID2, delay) # changed in v1.5.x
                grad_data[1] *= modifier  # first # changed in v1.5.x
                grad_data[2] *= modifier  # last # changed in v1.5.x

            # Use EventLibrary.update() to properly maintain keymap integrity
            new_data = tuple(grad_data)
            self.grad_library.update(event_id, old_data, new_data, grad_type)

        # Clear block cache to ensure get_block() uses the modified gradient data
        if self.use_block_cache:
            self.block_cache.clear()

    def paper_plot(
        self,
        time_range: Tuple[float] = (0, np.inf),
        line_width: float = 1.2,
        axes_color: Tuple[float] = (0.5, 0.5, 0.5),
        rf_color: str = 'black',
        gx_color: str = 'blue',
        gy_color: str = 'red',
        gz_color: Tuple[float] = (0, 0.5, 0.3),
        rf_plot: str = 'abs',
    ):
        """
        Plot sequence using paper-style formatting (minimalist, high-contrast layout).

        Parameters
        ----------
        time_range : iterable, default=(0, np.inf)
            Time range (x-axis limits) for plotting the sequence.
            Default is 0 to infinity (entire sequence).
        line_width : float, default=1.2
            Line width used in plots.
        axes_color : color, default=(0.5, 0.5, 0.5)
            Color of horizontal zero axes (e.g., gray).
        rf_color : color, default='black'
            Color for RF and ADC events.
        gx_color : color, default='blue'
            Color for gradient X waveform.
        gy_color : color, default='red'
            Color for gradient Y waveform.
        gz_color : color, default=(0, 0.5, 0.3)
            Color for gradient Z waveform.
        rf_plot : {'abs', 'real', 'imag'}, default='abs'
            Determines how to plot RF waveforms (magnitude, real or imaginary part).

        """
        ext_paper_plot(self, time_range, line_width, axes_color, rf_color, gx_color, gy_color, gz_color, rf_plot)

    def plot(
        self,
        label: str = str(),
        show_blocks: bool = False,
        save: bool = False,
        time_range=(0, np.inf),
        time_disp: str = 's',
        grad_disp: str = 'kHz/m',
        plot_now: bool = True,
        clear: bool = True,
        overlay: SeqPlot = None,
        stacked: bool = False,
        show_guides: bool = False,
    ) -> SeqPlot:
        """
        Plot `Sequence`.

        Parameters
        ----------
        label : str, default=str()
            Plot label values for ADC events: in this example for LIN and REP labels; other valid labels are accepted as
            a comma-separated list.
        show_blocks : bool, default=False
            Boolean flag to indicate if grid and tick labels at the block boundaries are to be plotted.
        save : bool, default=False
            Boolean flag indicating if plots should be saved. The two figures will be saved as JPG with numerical
            suffixes to the filename 'seq_plot'.
        time_range : iterable, default=(0, np.inf)
            Time range (x-axis limits) for plotting the sequence. Default is 0 to infinity (entire sequence).
        time_disp : str, default='s'
            Time display type, must be one of 's', 'ms' or 'us'.
        grad_disp : str, default='kHz/m'
            Gradient display unit, must be one of 'kHz/m' or 'mT/m'.
        plot_now : bool, default=True
            If True, function immediately shows the plots, blocking the rest of the code until plots are exited.
            If False, plots are shown when plt.show() is called. Useful if plots are to be modified.
        clear : bool, default=True
            If True, clear existing figures before plotting (default behavior).
            If False, overlay on existing figures 1 and 2 for sequence comparison.
        overlay : SeqPlot or None, default=None
            If provided, overlay this plot on the figures from the given SeqPlot object. Overrides fig1, fig2, and sets clear=False.
        stacked : bool, default=False
            If True, plot all channels (ADC, RF mag, RF phase, Gx, Gy, Gz) in a single stacked figure (MATLAB Pulseq style).
            If False, use separate figures for RF/ADC and gradients.
        show_guides : bool, default=False
            If True, enable dynamic vertical hairline guides that follow the cursor. Requires `mplcursors`.

        Returns
        -------
        SeqPlot
            SeqPlot handle.
        """
        return SeqPlot(
            self,
            label,
            show_blocks,
            save,
            time_range,
            time_disp,
            grad_disp,
            plot_now,
            clear,
            overlay,
            stacked,
            show_guides,
        )

    def read(self, file_path: str, detect_rf_use: bool = False, remove_duplicates: bool = True) -> None:
        """
        Read `.seq` file from `file_path`.

        Parameters
        ----------
        detect_rf_use
        file_path : str
            Path to `.seq` file to be read.
        remove_duplicates : bool, default=True
            Remove duplicate events from the sequence after reading.
        """
        if self.use_block_cache:
            self.block_cache.clear()

        read(self, path=file_path, detect_rf_use=detect_rf_use, remove_duplicates=remove_duplicates)

        # Initialize next free block ID
        self.next_free_block_ID = (max(self.block_events) + 1) if self.block_events else 1

    def register_adc_event(self, event: EventLibrary) -> int:
        return block.register_adc_event(self, event)

    def register_grad_event(self, event: SimpleNamespace) -> Union[int, Tuple[int, int]]:
        return block.register_grad_event(self, event)

    def register_label_event(self, event: SimpleNamespace) -> int:
        return block.register_label_event(self, event)

    def register_rf_event(self, event: SimpleNamespace) -> Tuple[int, List[int]]:
        return block.register_rf_event(self, event)

    def register_soft_delay_event(self, event: SimpleNamespace) -> int:
        return block.register_soft_delay_event(self, event)

    def remove_duplicates(self, in_place: bool = False) -> 'Sequence':
        """
        Removes duplicate events from the shape and event libraries contained
        in this sequence.

        Parameters
        ----------
        in_place : bool, optional
            If true, removes the duplicates from the current sequence.
            Otherwise, a copy is created. The default is False.

        Returns
        -------
        seq_copy : Sequence
            If `in_place`, returns self. Otherwise returns a copy of the
            sequence.
        """
        if in_place:
            seq_copy = self
        else:
            # Avoid copying block_cache for performance
            tmp = self.block_cache
            self.block_cache = {}
            seq_copy = deepcopy(self)
            self.block_cache = tmp

        # Find duplicate in shape library
        seq_copy.shape_library, mapping = seq_copy.shape_library.remove_duplicates(9)

        # Remap shape IDs of arbitrary gradient events
        for grad_id in seq_copy.grad_library.data:
            if seq_copy.grad_library.type[grad_id] == 'g':
                data = seq_copy.grad_library.data[grad_id]
                new_data = (*data[0:3], mapping[data[3]], mapping[data[4]], data[5])
                if data != new_data:
                    seq_copy.grad_library.update(grad_id, None, new_data)

        # Remap shape IDs of RF events
        for rf_id in seq_copy.rf_library.data:
            data = seq_copy.rf_library.data[rf_id]
            new_data = (data[0], mapping[data[1]], mapping[data[2]], mapping[data[3]], *data[4:])
            if data != new_data:
                seq_copy.rf_library.update(rf_id, None, new_data)

        # Filter duplicates in gradient library
        seq_copy.grad_library, mapping = seq_copy.grad_library.remove_duplicates((6, -6, -6, -6, -6, -6))

        # Remap gradient event IDs
        for block_id in seq_copy.block_events:
            seq_copy.block_events[block_id][2] = mapping[seq_copy.block_events[block_id][2]]
            seq_copy.block_events[block_id][3] = mapping[seq_copy.block_events[block_id][3]]
            seq_copy.block_events[block_id][4] = mapping[seq_copy.block_events[block_id][4]]

        # Filter duplicates in RF library
        seq_copy.rf_library, mapping = seq_copy.rf_library.remove_duplicates((6, 0, 0, 0, 6, 6, 6, 6, 6, 6))

        # Remap RF event IDs
        for block_id in seq_copy.block_events:
            seq_copy.block_events[block_id][1] = mapping[seq_copy.block_events[block_id][1]]

        # Filter duplicates in ADC library
        seq_copy.adc_library, mapping = seq_copy.adc_library.remove_duplicates((0, -9, -6, 6, 6, 6, 6, 6, 6))

        # Remap ADC event IDs
        for block_id in seq_copy.block_events:
            seq_copy.block_events[block_id][5] = mapping[seq_copy.block_events[block_id][5]]

        return seq_copy

    def rf_from_lib_data(self, lib_data: list, use: str = '') -> SimpleNamespace:
        """
        Construct RF object from `lib_data`.

        Parameters
        ----------
        lib_data : list
            RF envelope.
        use : str, default=''
            RF event use.

        Returns
        -------
        rf : SimpleNamespace
            RF object constructed from `lib_data`.
        """
        rf = SimpleNamespace()
        rf.type = 'rf'

        amplitude, mag_shape, phase_shape = lib_data[0], lib_data[1], lib_data[2]
        shape_data = self.shape_library.data[mag_shape]
        compressed = SimpleNamespace()
        compressed.num_samples = shape_data[0]
        compressed.data = shape_data[1:]
        mag = decompress_shape(compressed)
        shape_data = self.shape_library.data[phase_shape]
        compressed.num_samples = shape_data[0]
        compressed.data = shape_data[1:]
        phase = decompress_shape(compressed)
        rf.signal = amplitude * mag * np.exp(1j * 2 * math.pi * phase)
        time_shape = lib_data[3]
        if time_shape > 0:
            shape_data = self.shape_library.data[time_shape]
            compressed.num_samples = shape_data[0]
            compressed.data = shape_data[1:]
            rf.t = decompress_shape(compressed) * self.rf_raster_time
            rf.shape_dur = math.ceil((rf.t[-1] - eps) / self.rf_raster_time) * self.rf_raster_time
        else:  # Generate default time raster on the fly
            rf.t = (np.arange(1, len(rf.signal) + 1) - 0.5) * self.rf_raster_time
            rf.shape_dur = len(rf.signal) * self.rf_raster_time

        rf.center = lib_data[4]  # v150: new field
        rf.delay = lib_data[5]  # v150: changed from lib_data[4] to lib_data[5]
        rf.freq_ppm = lib_data[6]  # v150: new field
        rf.phase_ppm = lib_data[7]  # v150: new field
        rf.freq_offset = lib_data[8]  # v150: changed from lib_data[5] to lib_data[8]
        rf.phase_offset = lib_data[9]  # v150: changed from lib_data[6] to lib_data[9]

        rf.dead_time = self.system.rf_dead_time
        rf.ringdown_time = self.system.rf_ringdown_time

        # TODO: fixme : use map built from pp.get_supported_rf_uses()
        use_cases = {
            'e': 'excitation',
            'r': 'refocusing',
            'i': 'inversion',
            's': 'saturation',
            'p': 'preparation',
            'o': 'other',
        }
        rf.use = use_cases.get(use, 'undefined')

        return rf

    def rf_times(
        self, time_range: Union[List[float], None] = None
    ) -> Tuple[List[float], np.ndarray, List[float], np.ndarray, np.ndarray]:
        """
        Return time points of excitations and refocusings.

        Returns
        -------
        t_excitation : List[float]
            Contains time moments of the excitation RF pulses
        fp_excitation : np.ndarray
            Contains frequency and phase offsets of the excitation RF pulses
        t_refocusing : List[float]
            Contains time moments of the refocusing RF pulses
        fp_refocusing : np.ndarray
            Contains frequency and phase offsets of the excitation RF pulses
        """
        # Collect RF timing data
        t_excitation = []
        fp_excitation = []
        t_refocusing = []
        fp_refocusing = []

        curr_dur = 0
        if time_range is None:
            blocks = self.block_events
        else:
            if len(time_range) != 2:
                raise ValueError('Time range must be list of two elements')
            if time_range[0] > time_range[1]:
                raise ValueError('End time of time_range must be after begin time')

            # Calculate end times of each block
            bd = np.array(list(self.block_durations.values()))
            t = np.cumsum(bd)
            # Search block end times for start of time range
            begin_block = np.searchsorted(t, time_range[0])
            # Search block begin times for end of time range
            end_block = np.searchsorted(t - bd, time_range[1], side='right')
            blocks = list(self.block_durations.keys())[begin_block:end_block]
            curr_dur = t[begin_block] - bd[begin_block]

        for block_counter in blocks:
            block = self.get_block(block_counter)

            if block.rf is not None:
                rf = block.rf

                tc = calc_rf_center(rf)[0]
                t = rf.delay + tc

                full_freq_offset = rf.freq_offset + rf.freq_ppm * 1e-6 * self.system.gamma * self.system.B0
                full_phase_offset = rf.phase_offset + rf.phase_ppm * 1e-6 * self.system.gamma * self.system.B0
                full_phase_offset = full_phase_offset + 2 * math.pi * full_freq_offset * tc

                if not hasattr(rf, 'use') or block.rf.use in [
                    'excitation',
                    'undefined',
                ]:
                    t_excitation.append(curr_dur + t)
                    fp_excitation.append([full_freq_offset, full_phase_offset])
                elif block.rf.use == 'refocusing':
                    t_refocusing.append(curr_dur + t)
                    fp_refocusing.append([full_freq_offset, full_phase_offset])

            curr_dur += self.block_durations[block_counter]

        if len(t_excitation) != 0:
            fp_excitation = np.array(fp_excitation).T
        else:
            fp_excitation = np.empty((2, 0))

        if len(t_refocusing) != 0:
            fp_refocusing = np.array(fp_refocusing).T
        else:
            fp_refocusing = np.empty((2, 0))

        return t_excitation, fp_excitation, t_refocusing, fp_refocusing

    def set_block(self, block_index: int, *args: SimpleNamespace) -> None:
        """
        Replace block at index with new block provided as block structure, add sequence block, or create a new block
        from events and store at position specified by index. The block or events are provided in uncompressed form and
        will be stored in the compressed, non-redundant internal libraries.

        See Also
        --------
        - `pypulseq.Sequence.sequence.Sequence.get_block()`
        - `pypulseq.Sequence.sequence.Sequence.add_block()`

        Parameters
        ----------
        block_index : int
            Index at which block is replaced.
        args : SimpleNamespace
            Block or events to be replaced/added or created at `block_index`.
        """
        if trace_enabled():
            self.block_trace[block_index] = SimpleNamespace(block=trace())

        block.set_block(self, block_index, *args)

        if block_index >= self.next_free_block_ID:
            self.next_free_block_ID = block_index + 1

    def set_definition(self, key: str, value: Union[float, int, list, np.ndarray, str, tuple]) -> None:
        """
        Modify a custom definition of the sequence. Set the user definition 'key' to value 'value'. If the definition
        does not exist it will be created.

        See also `pypulseq.Sequence.sequence.Sequence.get_definition()`.

        Parameters
        ----------
        key : str
            Definition key.
        value : int, list, np.ndarray, str or tuple
            Definition value.
        """
        if key == 'FOV' and np.max(value) > 1:
            text = 'Definition FOV uses values exceeding 1 m. '
            text += 'New Pulseq interpreters expect values in units of meters.'
            warn(text)

        self.definitions[key] = value

    def set_extension_string_ID(self, extension_str: str, extension_id: int) -> None:
        """
        Set numeric ID for the given string extension ID.

        Parameters
        ----------
        extension_str : str
            Given string extension ID.
        extension_id : int
            Given numeric extension ID.

        Raises
        ------
        ValueError
            If given numeric or string extension ID is not unique.
        """
        if extension_str in self.extension_string_idx or extension_id in self.extension_numeric_idx:
            raise ValueError('Numeric or string ID is not unique')

        self.extension_numeric_idx.append(extension_id)
        self.extension_string_idx.append(extension_str)
        assert len(self.extension_numeric_idx) == len(self.extension_string_idx)

    def apply_soft_delay(self, **kwargs):
        """
        Apply soft delay values to modify block durations in the sequence.

        This method updates the durations of blocks containing soft delay events
        based on the provided values. The new block duration is calculated using
        the formula: duration = (input_value / factor) + offset.

        Parameters
        ----------
        **kwargs : dict
            Keyword arguments where keys are soft delay hint strings and values
            are the desired delay values in seconds. Not all soft delays in the
            sequence need to be specified.

        Raises
        ------
        ValueError
            If a specified soft delay hint does not exist in the sequence.
        ValueError
            If the calculated block duration would be negative.
        ValueError
            If soft delay hint and numeric ID mapping is inconsistent.

        Examples
        --------
        Apply single soft delay:

        >>> seq.apply_soft_delay(TE=40e-3)  # Set TE to 40ms

        Apply multiple soft delays:

        >>> seq.apply_soft_delay(TE=50e-3, TR=2.0)  # Set TE to 50ms, TR to 2s

        See Also
        --------
        pypulseq.make_soft_delay : Create soft delay events

        Notes
        -----
        - Only soft delays present in the sequence can be modified
        - Block durations are automatically rounded to the block duration raster
        - A warning is issued if substantial rounding occurs
        """

        # Go through all the blocks and update durations, at the same time checking the consistency of the soft delays
        sd_str2numID = {}
        sd_numID2hint = {}
        sd_warns = {}
        for block_counters in self.block_events:
            block = self.get_block(block_counters)
            if block.soft_delay is not None:
                # Check the numeric ID consistency
                if block.soft_delay.hint not in sd_str2numID:
                    sd_str2numID[block.soft_delay.hint] = block.soft_delay.numID
                else:
                    if sd_str2numID[block.soft_delay.hint] != block.soft_delay.numID:
                        raise ValueError(
                            f"Soft delay in block {block_counters} with numeric ID {block.soft_delay.numID} and string hint '{block.soft_delay.hint}' is inconsistent with the previous occurrences of the same string hint"
                        )

                if block.soft_delay.numID not in sd_numID2hint:
                    sd_numID2hint[block.soft_delay.numID] = block.soft_delay.hint
                else:
                    if sd_numID2hint[block.soft_delay.numID] != block.soft_delay.hint:
                        raise ValueError(
                            f"Soft delay in block {block_counters} with numeric ID {block.soft_delay.numID} and string hint '{block.soft_delay.hint}' is inconsistent with the previous occurrences of the same numeric ID"
                        )

                if block.soft_delay.hint in kwargs:
                    # Calculate the new block duration
                    new_dur_ru = (
                        kwargs[block.soft_delay.hint] / block.soft_delay.factor + block.soft_delay.offset
                    ) / self.system.block_duration_raster
                    new_dur = round(new_dur_ru) * self.system.block_duration_raster
                    # Check if rounding error is significant (threshold: 0.5 microseconds)
                    rounding_threshold = 0.5e-6
                    rounding_error = abs(new_dur - new_dur_ru * self.system.block_duration_raster)
                    if rounding_error > rounding_threshold and block.soft_delay.numID not in sd_warns:
                        warn(
                            f"Soft delay '{block.soft_delay.hint}' in block {block_counters}: "
                            f'Duration rounded by {rounding_error * 1e6:.1f} μs to align with raster time '
                            f'({self.system.block_duration_raster * 1e6:.1f} μs). '
                            f'This warning is shown only once per soft delay ID.'
                        )
                        sd_warns[block.soft_delay.numID] = True

                    if new_dur < 0:
                        raise ValueError(
                            f"Soft delay '{block.soft_delay.hint}' in block {block_counters}: "
                            f'Calculated duration is negative ({new_dur * 1e6:.1f} μs). '
                            f'Check the offset ({block.soft_delay.offset * 1e6:.1f} μs) and factor '
                            f'({block.soft_delay.factor}) parameters.'
                        )

                    self.block_durations[block_counters] = new_dur

        # Now check if there are some input soft delays which haven't been found in the sequence
        all_input_hints = kwargs.keys()
        for hint in all_input_hints:
            if hint not in sd_str2numID:
                available_hints = list(sd_str2numID.keys())
                raise ValueError(
                    f"Soft delay '{hint}' not found in sequence. "
                    f'Available soft delays: {available_hints if available_hints else "none"}'
                )

    def test_report(self) -> str:
        """
        Analyze the sequence and return a text report.
        """
        return ext_test_report(self)

    def waveforms(self, append_RF: bool = False, time_range: Union[List[float], None] = None) -> Tuple[np.ndarray]:
        """
        Decompress the entire gradient waveform. Returns gradient waveforms as a tuple of `np.ndarray` of
        `gradient_axes` (typically 3) dimensions. Each `np.ndarray` contains timepoints and the corresponding
        gradient amplitude values.

        Parameters
        ----------
        append_RF : bool, default=False
            Boolean flag to indicate if RF wave shapes are to be appended after the gradients.

        Returns
        -------
        wave_data : np.ndarray
        """
        grad_channels = ['gx', 'gy', 'gz']

        # Collect shape pieces
        if append_RF:
            shape_channels = len(grad_channels) + 1  # Last 'channel' is RF
        else:
            shape_channels = len(grad_channels)

        shape_pieces = [[] for _ in range(shape_channels)]
        out_len = np.zeros(shape_channels)  # Last 'channel' is RF

        curr_dur = 0
        if time_range is None:
            blocks = self.block_events
        else:
            if len(time_range) != 2:
                raise ValueError('Time range must be list of two elements')
            if time_range[0] > time_range[1]:
                raise ValueError('End time of time_range must be after begin time')

            # Calculate end times of each block
            bd = np.array(list(self.block_durations.values()))
            t = np.cumsum(bd)
            # Search block end times for start of time range
            begin_block = np.searchsorted(t, time_range[0])
            # Search block begin times for end of time range
            end_block = np.searchsorted(t - bd, time_range[1], side='right')
            blocks = list(self.block_durations.keys())[begin_block:end_block]
            curr_dur = t[begin_block] - bd[begin_block]

        for block_counter in blocks:
            block = self.get_block(block_counter)

            for j in range(len(grad_channels)):
                grad = getattr(block, grad_channels[j])
                if grad is not None:  # Gradients
                    if grad.type == 'grad':
                        # Check if we have an extended trapezoid or an arbitrary gradient on a regular raster
                        tt_rast = grad.tt / self.grad_raster_time + 0.5
                        if np.all(np.abs(tt_rast - np.arange(1, len(tt_rast) + 1)) < eps):  # Arbitrary gradient
                            """
                            Arbitrary gradient: restore & recompress shape - if we had a trapezoid converted to shape we
                            have to find the "corners" and we can eliminate internal samples on the straight segments
                            but first we have to restore samples on the edges of the gradient raster intervals for that
                            we need the first sample.
                            """

                            # TODO: Implement restoreAdditionalShapeSamples
                            #       https://github.com/pulseq/pulseq/blob/master/matlab/%2Bmr/restoreAdditionalShapeSamples.m

                            out_len[j] += len(grad.tt) + 2
                            shape_pieces[j].append(
                                np.array(
                                    [
                                        curr_dur
                                        + grad.delay
                                        + np.concatenate(([0], grad.tt, [grad.tt[-1] + self.grad_raster_time / 2])),
                                        np.concatenate(([grad.first], grad.waveform, [grad.last])),
                                    ]
                                )
                            )
                        else:  # Extended trapezoid
                            out_len[j] += len(grad.tt)
                            shape_pieces[j].append(
                                np.array(
                                    [
                                        curr_dur + grad.delay + grad.tt,
                                        grad.waveform,
                                    ]
                                )
                            )
                    else:
                        if abs(grad.flat_time) > eps:
                            out_len[j] += 4
                            _temp = np.vstack(
                                (
                                    cumsum(
                                        curr_dur + grad.delay,
                                        grad.rise_time,
                                        grad.flat_time,
                                        grad.fall_time,
                                    ),
                                    grad.amplitude * np.array([0, 1, 1, 0]),
                                )
                            )
                            shape_pieces[j].append(_temp)
                        else:
                            if abs(grad.rise_time) > eps and abs(grad.fall_time) > eps:
                                out_len[j] += 3
                                _temp = np.vstack(
                                    (
                                        cumsum(curr_dur + grad.delay, grad.rise_time, grad.fall_time),
                                        grad.amplitude * np.array([0, 1, 0]),
                                    )
                                )
                                shape_pieces[j].append(_temp)
                            else:
                                if abs(grad.amplitude) > eps:
                                    print(
                                        'Warning: "empty" gradient with non-zero magnitude detected in block {}'.format(
                                            block_counter
                                        )
                                    )

            if block.rf is not None:  # RF
                rf = block.rf
                full_freq_offset = rf.freq_offset + rf.freq_ppm * 1e-6 * self.system.gamma * self.system.B0
                full_phase_offset = rf.phase_offset + rf.phase_ppm * 1e-6 * self.system.gamma * self.system.B0
                if append_RF:
                    rf_piece = np.array(
                        [
                            curr_dur + rf.delay + rf.t,
                            rf.signal * np.exp(1j * (full_phase_offset + 2 * math.pi * full_freq_offset * rf.t)),
                        ]
                    )
                    out_len[-1] += len(rf.t)

                    if abs(rf.signal[0]) > 0:
                        pre = np.array([[rf_piece[0, 0] - 0.1 * self.system.rf_raster_time], [0]])
                        rf_piece = np.hstack((pre, rf_piece))
                        out_len[-1] += pre.shape[1]

                    if abs(rf.signal[-1]) > 0:
                        post = np.array([[rf_piece[0, -1] + 0.1 * self.system.rf_raster_time], [0]])
                        rf_piece = np.hstack((rf_piece, post))
                        out_len[-1] += post.shape[1]

                    shape_pieces[-1].append(rf_piece)

            curr_dur += self.block_durations[block_counter]

        # Collect wave data
        wave_data = []

        for j in range(shape_channels):
            if shape_pieces[j] == []:
                wave_data.append(np.zeros((2, 0)))
                continue

            # If the first element of the next shape has the same time as
            # the last element of the previous shape, drop the first
            # element of the next shape.
            shape_pieces[j] = [shape_pieces[j][0]] + [
                cur if prev[0, -1] + eps < cur[0, 0] else cur[:, 1:]
                for prev, cur in zip(shape_pieces[j][:-1], shape_pieces[j][1:], strict=False)
            ]

            wave_data.append(np.concatenate(shape_pieces[j], axis=1))

            rftdiff = np.diff(wave_data[j][0])
            if np.any(rftdiff < eps):
                raise Warning('Time vector elements are not monotonically increasing.')

        return wave_data

    def waveforms_and_times(
        self, append_RF: bool = False, time_range: Union[List[float], None] = None
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Decompress the entire gradient waveform. Returns gradient waveforms as a tuple of `np.ndarray` of
        `gradient_axes` (typically 3) dimensions. Each `np.ndarray` contains timepoints and the corresponding
        gradient amplitude values. Additional return values are time points of excitations, refocusings and ADC
        sampling points.

        Parameters
        ----------
        append_RF : bool, default=False
            Boolean flag to indicate if RF wave shapes are to be appended after the gradients.

        Returns
        -------
        wave_data : np.ndarray
        tfp_excitation : np.ndarray
            Contains time moments, frequency and phase offsets of the excitation RF pulses (similar for `
            tfp_refocusing`).
        tfp_refocusing : np.ndarray
        t_adc: np.ndarray
            Contains times of all ADC sample points.
        fp_adc : np.ndarray
            Contains frequency and phase offsets of each ADC object (not samples).
        """
        wave_data = self.waveforms(append_RF=append_RF, time_range=time_range)
        t_excitation, fp_excitation, t_refocusing, fp_refocusing = self.rf_times(time_range=time_range)
        t_adc, fp_adc = self.adc_times(time_range=time_range)

        # Join times, frequency and phases of RF pulses for compatibility with previous implementation
        tfp_excitation = np.concatenate((np.array(t_excitation)[None], fp_excitation), axis=0)
        tfp_refocusing = np.concatenate((np.array(t_refocusing)[None], fp_refocusing), axis=0)

        return wave_data, tfp_excitation, tfp_refocusing, t_adc, fp_adc

    def write(
        self,
        name: str,
        create_signature: bool = True,
        remove_duplicates: bool = True,
        check_timing: bool = True,
        v141_compat: bool = False,
    ) -> Union[str, None]:
        """
        Write the sequence data to the given filename using the open file format for MR sequences.

        See also `pypulseq.Sequence.read_seq.read()`.

        Parameters
        ----------
        name : str
            Filename of `.seq` file to be written to disk.
        create_signature : bool, default=True
            Boolean flag to indicate if the file has to be signed.
        remove_duplicates : bool, default=True
            Remove duplicate events from the sequence before writing
        v141_compat: bool, default=False
            Write the sequence in v1.4.1 compatible file format.

        Returns
        -------
        signature or None : If create_signature is True, it returns the written .seq file's signature as a string,
        otherwise it returns None. Note that, if remove_duplicates is True, signature belongs to the
        deduplicated sequences signature, and not the Sequence that is stored in the Sequence object.
        """
        # Check if there are any timing errors in the sequence
        if check_timing:
            is_ok, error_report = self.check_timing()
            if not is_ok:
                warn(f'write(): {len(error_report)} timing errors found in the sequence', stacklevel=2)

        # Calculate sequence duration and stored it in the TotalDuration definition
        self.set_definition('TotalDuration', sum(self.block_durations.values()))

        # Check whether all gradients in the last block are ramped down properly
        last_block_id = next(reversed(self.block_events))
        last_block = self.get_block(last_block_id)
        for channel, event in zip(('x', 'y', 'z'), (last_block.gx, last_block.gy, last_block.gz), strict=False):
            if (
                event is not None
                and event.type == 'grad'
                and abs(event.last) > self.system.max_slew * self.system.grad_raster_time
            ):
                warn_msg = f'write(): Gradient on channel {channel} in last sequence block does not ramp down to 0'

                if trace_enabled():
                    trace = self.block_trace.get(last_block_id, None)

                    if hasattr(trace, 'block'):
                        warn_msg += '\nLast block defined here:\n' + format_trace(trace.block)
                    if hasattr(trace, 'g' + channel):
                        warn_msg += f'\n`g{channel}` defined here:\n' + format_trace(getattr(trace, 'g' + channel))

                warn(warn_msg, stacklevel=2)

        # Write the sequence
        if v141_compat:
            signature = write_seq_v141(self, name, create_signature, remove_duplicates)
        else:
            signature = write_seq(self, name, create_signature, remove_duplicates)

        # Return the sequence md5 signature if requested
        if signature is not None:
            self.signature_type = 'md5'
            self.signature_file = 'text'
            self.signature_value = signature
            return signature
        else:
            return None



from types import SimpleNamespace
from typing import Any, List, Tuple

from pypulseq import Sequence, eps
from pypulseq.calc_duration import calc_duration
from pypulseq.utils.tracing import format_trace

error_messages = {
    'RASTER': '{value*multiplier:.2f} {unit} does not align to {raster} (Nearest valid value: {value_rounded*multiplier:.0f} {unit}, error: {error*multiplier:.2f} {unit})',
    'ADC_DEAD_TIME': 'ADC delay is smaller than ADC dead time ({value*multiplier:.2f} {unit} < {dead_time*multiplier:.0f} {unit})',
    'POST_ADC_DEAD_TIME': 'Post-ADC dead time exceeds block duration ({value*multiplier:.2f} {unit} + {dead_time*multiplier:.0f} {unit} > {duration*multiplier} {unit})',
    'BLOCK_DURATION_MISMATCH': 'Inconsistency between the stored block duration ({duration*multiplier:.2f} {unit}) and the content of the block ({value*multiplier:.2f} {unit})',
    'RF_DEAD_TIME': 'Delay of {value*multiplier:.2f} {unit} is smaller than the RF dead time {dead_time*multiplier:.0f} {unit}',
    'RF_RINGDOWN_TIME': 'Time between the end of the RF pulse at {value*multiplier:.2f} {unit} and the end of the block at {duration * multiplier:.2f} {unit} is shorter than rf_ringdown_time ({ringdown_time*multiplier:.0f} {unit})',
    'NEGATIVE_DELAY': 'Delay is negative {value*multiplier:.2f} {unit}',
    'SOFT_DELAY_FACTOR': 'Soft delay {hint}/{numID} has factor parameter as zero, which makes duration calculation undefined.',
    'SOFT_DELAY_DUR_INCONSISTENCY': 'Soft delay {hint}/{numID} default duration derived from this block ({value*1e6} us) is inconsistent with the previous default.',
    'SOFT_DELAY_HINT_INCONSISTENCY': 'Soft delay {hint}/{numID}: Soft delays with the same numeric ID are expected to share the same text hint but previous hint recorded is {prev_hint}.',
    'SOFT_DELAY_INVALID_NUMID': 'Soft delay {hint}/{numID} has an invalid numeric ID {numID}. Numeric IDs must be positive integers.',
}


def check_timing(seq: Sequence) -> Tuple[bool, List[SimpleNamespace]]:
    error_report: List[SimpleNamespace] = []

    def div_check(a: float, b: float, event: str, field: str, raster: str):
        """
        Checks whether `a` can be divided by `b` to an accuracy of 1e-9.
        """
        c = a / b
        c_rounded = round(c)
        is_ok = abs(c - c_rounded) < 1e-6

        if not is_ok:
            error_report.append(
                SimpleNamespace(
                    block=block_counter,
                    event=event,
                    field=field,
                    value=a,
                    value_rounded=c_rounded * b,
                    error=(a - c_rounded * b),
                    raster=raster,
                    error_type='RASTER',
                )
            )

    soft_delay_defaults = {}

    # Loop over all blocks
    for block_counter in seq.block_events:
        block = seq.get_block(block_counter)

        # Check block duration
        duration = calc_duration(block)
        div_check(
            duration, seq.system.block_duration_raster, event='block', field='duration', raster='block_duration_raster'
        )

        if abs(duration - seq.block_durations[block_counter]) > eps:
            error_report.append(
                SimpleNamespace(
                    block=block_counter,
                    event='block',
                    field='duration',
                    error_type='BLOCK_DURATION_MISMATCH',
                    value=duration,
                    duration=seq.block_durations[block_counter],
                )
            )
            duration = seq.block_durations[block_counter]

        # Check block events
        for event, e in block.__dict__.items():
            if e is None or isinstance(e, (float, int)):  # Special handling for block_duration
                continue
            elif not isinstance(e, (dict, SimpleNamespace)):
                raise ValueError('Wrong data type of variable arguments, list[SimpleNamespace] expected.')

            if isinstance(e, list) and len(e) > 1:
                # For now this is only the case for arrays of extensions, but we cannot actually check extensions anyway...
                continue

            if hasattr(e, 'type') and e.type == 'rf':
                raster = seq.system.rf_raster_time
                raster_str = 'rf_raster_time'
            elif hasattr(e, 'type') and e.type == 'adc':
                # note that ADC samples must be on ADC raster time, but the ADC start time must be on RF raster time!
                # see https://github.com/pulseq/pulseq/blob/master/doc%2Fpulseq_shapes_and_times.pdf for details
                raster = seq.system.rf_raster_time
                raster_str = 'rf_raster_time'
            else:
                raster = seq.system.grad_raster_time
                raster_str = 'grad_raster_time'

            if hasattr(e, 'delay'):
                if e.delay < -eps:
                    error_report.append(
                        SimpleNamespace(
                            block=block_counter, event=event, field='delay', error_type='NEGATIVE_DELAY', value=e.delay
                        )
                    )

                div_check(e.delay, raster, event=event, field='delay', raster=raster_str)

            if hasattr(e, 'duration'):
                div_check(e.duration, raster, event=event, field='duration', raster=raster_str)

            if hasattr(e, 'dwell'):
                div_check(e.dwell, seq.system.adc_raster_time, event=event, field='dwell', raster='adc_raster_time')

            if hasattr(e, 'type') and e.type == 'trap':
                div_check(
                    e.rise_time, seq.system.grad_raster_time, event=event, field='rise_time', raster='grad_raster_time'
                )
                div_check(
                    e.flat_time, seq.system.grad_raster_time, event=event, field='flat_time', raster='grad_raster_time'
                )
                div_check(
                    e.fall_time, seq.system.grad_raster_time, event=event, field='fall_time', raster='grad_raster_time'
                )

        # Check RF dead times
        if block.rf is not None:
            if block.rf.delay - block.rf.dead_time < -eps:
                error_report.append(
                    SimpleNamespace(
                        block=block_counter,
                        event='rf',
                        field='delay',
                        error_type='RF_DEAD_TIME',
                        value=block.rf.delay,
                        dead_time=block.rf.dead_time,
                    )
                )

            if block.rf.delay + block.rf.t[-1] + block.rf.ringdown_time - duration > eps:
                error_report.append(
                    SimpleNamespace(
                        block=block_counter,
                        event='rf',
                        field='duration',
                        error_type='RF_RINGDOWN_TIME',
                        value=block.rf.delay + block.rf.t[-1],
                        duration=duration,
                        ringdown_time=block.rf.ringdown_time,
                    )
                )

        # Check ADC dead times
        if block.adc is not None:
            if block.adc.delay - seq.system.adc_dead_time < -eps:
                error_report.append(
                    SimpleNamespace(
                        block=block_counter,
                        event='adc',
                        field='delay',
                        error_type='ADC_DEAD_TIME',
                        value=block.adc.delay,
                        dead_time=seq.system.adc_dead_time,
                    )
                )

            adc_end = block.adc.delay + block.adc.num_samples * block.adc.dwell + seq.system.adc_dead_time

            if adc_end > duration + eps:
                error_report.append(
                    SimpleNamespace(
                        block=block_counter,
                        event='adc',
                        field='duration',
                        error_type='POST_ADC_DEAD_TIME',
                        value=block.adc.delay + block.adc.num_samples * block.adc.dwell,
                        duration=duration,
                        dead_time=seq.system.adc_dead_time,
                    )
                )
        if block.soft_delay is not None:
            if block.soft_delay.factor == 0:
                error_report.append(
                    SimpleNamespace(
                        block=block_counter,
                        event='soft_delay',
                        field='delay',
                        error_type='SOFT_DELAY_FACTOR',
                        value=block.soft_delay.factor,
                        hint=block.soft_delay.hint,
                        numID=block.soft_delay.numID,
                    )
                )
            # Calculate default delay based on the current block duration
            default_delay = (seq.block_durations[block_counter] - block.soft_delay.offset) * block.soft_delay.factor
            if block.soft_delay.numID not in soft_delay_defaults:
                soft_delay_defaults[block.soft_delay.numID] = default_delay
            elif (
                abs(default_delay - soft_delay_defaults[block.soft_delay.numID])
                > 1e-7  # 0.1 μs threshold for duration consistency
            ):
                error_report.append(
                    SimpleNamespace(
                        block=block_counter,
                        event='soft_delay',
                        field='delay',
                        error_type='SOFT_DELAY_DUR_INCONSISTENCY',
                        value=default_delay,
                        hint=block.soft_delay.hint,
                        numID=block.soft_delay.numID,
                    )
                )

    return len(error_report) == 0, error_report


def format_string(template: str, **kwargs: Any) -> str:
    """
    Evaluate a formatted string using the f-string syntax. Similar to
    `'{x}'.format(x=1)`, but allows arbitrary computations, e.g.
    `'{x*y}'.format(x=2,y=2)`.

    Parameters
    ----------
    template : str
        Format string.
    **kwargs : Any
        Variables to use in the formatted string.

    Returns
    -------
    str
        Formatted string.
    """
    return eval(f'f"""{template}"""', kwargs)


def indent_string(x: str, n: int = 2) -> str:
    """
    Adds indentations (`n` spaces) to every line in a string
    """
    return '\n'.join(' ' * n + y for y in x.splitlines())


def print_error_report(
    seq: Sequence,
    error_report: List[SimpleNamespace],
    full_report: bool = False,
    max_errors: int = 10,
    colored: bool = True,
) -> None:
    current_block = None

    if full_report:
        max_errors = len(error_report)

    for e in error_report[:max_errors]:
        if e.block != current_block:
            print(f'Block {e.block}:')
            current_block = e.block
            trace = seq.block_trace.get(current_block, None)

            if hasattr(trace, 'block'):
                print(
                    ('\x1b[38;5;8m' if colored else '')
                    + 'Block created here:\n'
                    + format_trace(trace.block)
                    + ('\x1b[0m' if colored else '')
                )

        unit = 'us'
        multiplier = 1e6
        if e.field == 'dwell':
            unit = 'ns'
            multiplier = 1e9

        error_message = format_string(error_messages[e.error_type], **e.__dict__, unit=unit, multiplier=multiplier)
        print(
            f'- {e.event}.{e.field}: '
            + ('\x1b[38;5;9m' if colored else '')
            + error_message
            + ('\x1b[0m' if colored else '')
        )

        if hasattr(trace, e.event) and e.event != 'block':
            print(
                ('\x1b[38;5;8m' if colored else '')
                + f'  `{e.event}` created here:\n'
                + format_trace(getattr(trace, e.event), indent=2)
                + ('\x1b[0m' if colored else '')
            )

    if len(error_report) > max_errors:
        blocks = [e.block for e in error_report[max_errors:]]

        print(f'--- {len(error_report) - max_errors} more errors in blocks {min(blocks)} to {max(blocks)} hidden ---')



import itertools
from math import ceil, floor, gcd, isclose, prod
from types import SimpleNamespace
from typing import List, Tuple, Union
from warnings import warn

import numpy as np

from pypulseq.opts import Opts
from pypulseq.utils.tracing import trace, trace_enabled


def make_adc(
    num_samples: int,
    delay: float = 0.0,
    duration: float = 0.0,
    dwell: float = 0.0,
    freq_offset: float = 0.0,
    phase_offset: float = 0.0,
    system: Union[Opts, None] = None,
    freq_ppm: float = 0.0,
    phase_ppm: float = 0.0,
    phase_modulation: Union[np.ndarray, None] = None,
) -> SimpleNamespace:
    """
    Create an ADC readout event.

    Parameters
    ----------
    num_samples: int
        Number of readout samples.
    system : Opts, default=Opts()
        System limits. Default is a system limits object initialized to default values.
    dwell : float, default=0
        ADC dead time in seconds (s) after sampling.
    duration : float, default=0
        Duration in seconds (s) of ADC readout event with `num_samples` number of samples.
    delay : float, default=0
        Delay in seconds (s) of ADC readout event.
    freq_offset : float, default=0
        Frequency offset of ADC readout event.
    phase_offset : float, default=0
        Phase offset of ADC readout event.
    freq_ppm : float, default=0
        PPM frequency offset of ADC readout event.
    phase_ppm : float, default=0
        PPM phase offset of ADC readout event.
    phase_modulation : Union[np.ndarray, None], default=None
        Phase modulation array for FOV shifting.
        If provided, it must have `num_samples` number of samples.

    Returns
    -------
    adc : SimpleNamespace
        ADC readout event.

    Raises
    ------
    ValueError
        If neither `dwell` nor `duration` are defined.
    """
    if system is None:
        system = Opts.default

    adc = SimpleNamespace()
    adc.type = 'adc'
    adc.num_samples = num_samples
    adc.dwell = dwell
    adc.delay = delay
    adc.freq_offset = freq_offset
    adc.phase_offset = phase_offset
    adc.freq_ppm = freq_ppm
    adc.phase_ppm = phase_ppm
    adc.dead_time = system.adc_dead_time

    if (dwell == 0 and duration == 0) or (dwell > 0 and duration > 0):
        raise ValueError('Either dwell or duration must be defined')

    if phase_modulation is not None and len(phase_modulation) != num_samples:
        raise ValueError('ADC Phase modulation vector must have the same length as the number of samples')
    adc.phase_modulation = phase_modulation

    if duration > 0:
        adc.dwell = duration / num_samples

    if dwell > 0:
        adc.duration = dwell * num_samples

    if adc.dead_time > adc.delay:
        warn(
            f'Specified ADC delay {adc.delay * 1e6:.2f} us is less than the dead time {adc.dead_time * 1e6:.0f} us. Delay was increased to the dead time.',
            stacklevel=2,
        )
        adc.delay = adc.dead_time

    if trace_enabled():
        adc.trace = trace()

    return adc


def calc_adc_segments(
    num_samples: int, dwell: float, system: Union[Opts, None] = None, mode: str = 'lengthen'
) -> Tuple[int, int]:
    """Calculate splitting of the ADC in segments with equal samples.

    Parameters
    ----------
    num_samples : int
        Initial number of samples to split into segments.
    dwell : float
        Dwell time of the ADC in [s]
    system : Opts, default=None
       System limits. Default is a system limits object initialized to default values.
    mode : str, default='lengthen'
        The total number of samples can either be shortened or lengthened to match the constraints.

    Returns
    -------
    (num_segments, num_samples_seg): int, int
        Number of segments and number of samples per segment.

    Notes
    -----
    "On some scanners, notably Siemens, ADC objects that exceed a certain
    sample length (8192 samples on Siemens) should be splittable to N
    equal parts, each of which aligned to the gradient raster. Each
    segment, however, needs to have the number of samples smaller than
    system.adcSamplesLimit' and divisible by system.adcSamplesDivisor to be
    executable on the scanner. The optional parameter mode can be either
    'shorten' or 'lengthen'." [Matlab implementation of 'calcAdcSeg'. https://github.com/pulseq]

    Raises
    ------
    ValueError
        If 'mode' is not 'shorten' or 'lengthen'.
    ValueError
        If no suitable segmentation could be found.
    ValueError
        If number of segments exceeds 128.
    """
    # Define maximum number of segments for the ADC
    max_segments = 128

    if mode not in ['shorten', 'lengthen']:
        raise ValueError(f"'mode' must be 'shorten' or 'lengthen' but is {mode}")

    if system is None:
        system = Opts.default

    # Check if single adc is sufficient
    if system.adc_samples_limit <= 0:
        return 1, num_samples

    # Get minimum number of samples for which the adc duration
    # is a multiple of grad raster time (GRT) and adc raster time (ART)
    i_gr = round(system.grad_raster_time / system.adc_raster_time)  # GRT in units of ART
    if not isclose(system.grad_raster_time / system.adc_raster_time, i_gr):
        raise ValueError("System 'grad_raster_time' is not a multiple of 'adc_raster_time'.")

    i_dwell = round(dwell / system.adc_raster_time)  # dwell in units of ART
    if not isclose(dwell / system.adc_raster_time, i_dwell):
        raise ValueError("'dwell' is not a multiple of system 'adc_raster_time'.")

    i_common = _lcm(i_gr, i_dwell)
    min_samples_segment = int(i_common / i_dwell)  # lcm(a,b)/b is always int

    # Siemens: Number of Samples should be divisible by a divisor
    gcd_adcdiv = gcd(min_samples_segment, system.adc_samples_divisor)
    if gcd_adcdiv != system.adc_samples_divisor:
        min_samples_segment *= system.adc_samples_divisor / gcd_adcdiv

    # Get segment multiplier
    if mode == 'shorten':
        samples_seg_multip = floor(num_samples / min_samples_segment)
    else:
        samples_seg_multip = ceil(num_samples / min_samples_segment)
    while samples_seg_multip > 0 and samples_seg_multip < (2 * num_samples / min_samples_segment):
        # Get prime factors from segments multiplier
        adc_seg_primes = _prime_factors(samples_seg_multip)
        num_segments = 1
        if len(adc_seg_primes) > 1:
            num_segments_candids = set()
            for k in range(1, len(adc_seg_primes) + 1):
                num_segments_candids |= {prod(perm) for perm in itertools.combinations(adc_seg_primes, k)}
            # Find suitable candidate
            for num_segments in sorted(num_segments_candids):
                num_samples_seg = samples_seg_multip * min_samples_segment / num_segments
                if num_samples_seg <= system.adc_samples_limit and num_segments <= max_segments:
                    break  # Found segments and samples
        else:  # Only one possible solution
            num_samples_seg = samples_seg_multip * min_samples_segment

        # Does output already fulfill constraints?
        if num_samples_seg <= system.adc_samples_limit and num_segments <= max_segments:
            break
        else:  # Shorten or lengthen the number of samples per segment
            samples_seg_multip += 1 if mode == 'lengthen' else -1

    # Validate constraints
    if samples_seg_multip <= 0:
        raise ValueError('Could not find suitable segmentation.')
    if num_samples_seg == 0:
        raise ValueError('Could not find suitable number of samples per segment.')
    if num_segments > max_segments:
        raise ValueError(f'Number of segments ({num_segments}) exceeds allowed number of {max_segments}')

    return int(num_segments), int(num_samples_seg)


def _prime_factors(n: int) -> List[int]:
    """Compute the prime factors of given integer n."""
    if n == 1:
        return [1]
    else:
        factors = []
        # Dividing n by 2 until it's odd
        while n % 2 == 0:
            factors.append(2)
            n //= 2

        # Checking odd factors from 3 upwards
        i = 3
        while i * i <= n:
            while n % i == 0:
                factors.append(i)
                n //= i
            i += 2
        if n > 2:
            factors.append(n)

        return factors


def _lcm(a, b):
    """Calculates the least common multiple of a and b."""
    return (a * b) // gcd(a, b)



from typing import Union

from pypulseq.convert import convert


class Opts:
    """
    System limits of an MR scanner.

    Note: Default values can be overwritten by creating an Opts object and
    calling `set_as_default`.

    Attributes
    ----------
    adc_dead_time : float, default=0
        Dead time for ADC readout pulses.
    adc_raster_time : float, default=100e-9
        Raster time for ADC readout pulses.
    block_duration_raster : float, default=10e-6
        Raster time for block durations.
    gamma : float, default=42.576e6
        Gyromagnetic ratio. Default gamma is specified for Hydrogen.
    grad_raster_time : float, default=10e-6
        Raster time for gradient waveforms.
    grad_unit : str, default='Hz/m'
        Unit of maximum gradient amplitude. Must be one of 'Hz/m', 'mT/m' or 'rad/ms/mm'.
    max_grad : float, default=40 mT/m
        Maximum gradient amplitude.
    max_slew : float, default=170 T/m/s
        Maximum slew rate.
    rf_dead_time : float, default=0
        Dead time for radio-frequency pulses.
    rf_raster_time : float, default=1e-6
        Raster time for radio-frequency pulses.
    rf_ringdown_time : float, default=0
        Ringdown time for radio-frequency pulses.
    adc_samples_limit : int, default=0
        Maximum number of samples for a single ADC object. If 0, no limit is set.
    adc_samples_divisor : int, default=4
        Samples of ADC must be divisible by 'adc_samples_divisor'.
    rise_time : float, default=0
        Rise time for gradients.
    slew_unit : str, default='Hz/m/s'
        Unit of maximum slew rate. Must be one of 'Hz/m/s', 'mT/m/ms', 'T/m/s' or 'rad/ms/mm/ms'.
    B0 : float, default=1.5
        Main magnetic field strength (in tesla)

    Raises
    ------
    ValueError
        If invalid `grad_unit` is passed. Must be one of 'Hz/m', 'mT/m' or 'rad/ms/mm'.
        If invalid `slew_unit` is passed. Must be one of 'Hz/m/s', 'mT/m/ms', 'T/m/s' or 'rad/ms/mm/ms'.
    """

    def __init__(
        self,
        adc_dead_time: Union[float, None] = None,
        adc_raster_time: Union[float, None] = None,
        block_duration_raster: Union[float, None] = None,
        gamma: Union[float, None] = None,
        grad_raster_time: Union[float, None] = None,
        grad_unit: str = 'Hz/m',
        max_grad: Union[float, None] = None,
        max_slew: Union[float, None] = None,
        rf_dead_time: Union[float, None] = None,
        rf_raster_time: Union[float, None] = None,
        rf_ringdown_time: Union[float, None] = None,
        adc_samples_limit: Union[int, None] = None,
        adc_samples_divisor: Union[int, None] = None,
        rise_time: Union[float, None] = None,
        slew_unit: str = 'Hz/m/s',
        B0: Union[float, None] = None,
    ):
        valid_grad_units = ['Hz/m', 'mT/m', 'rad/ms/mm']
        valid_slew_units = ['Hz/m/s', 'mT/m/ms', 'T/m/s', 'rad/ms/mm/ms']

        if grad_unit not in valid_grad_units:
            raise ValueError(
                f"Invalid gradient unit. Must be one of 'Hz/m', 'mT/m' or 'rad/ms/mm'. Passed: {grad_unit}"
            )

        if slew_unit not in valid_slew_units:
            raise ValueError(
                f"Invalid slew rate unit. Must be one of 'Hz/m/s', 'mT/m/ms', 'T/m/s' or 'rad/ms/mm/ms'. "
                f'Passed: {slew_unit}'
            )

        if gamma is None:
            gamma = Opts.default.gamma

        if max_grad is not None:
            max_grad = convert(from_value=max_grad, from_unit=grad_unit, to_unit='Hz/m', gamma=abs(gamma))
        else:
            max_grad = Opts.default.max_grad

        if max_slew is not None:
            max_slew = convert(from_value=max_slew, from_unit=slew_unit, to_unit='Hz/m', gamma=abs(gamma))
        else:
            max_slew = Opts.default.max_slew

        if rise_time is not None:
            max_slew = max_grad / rise_time

        if adc_dead_time is None:
            adc_dead_time = Opts.default.adc_dead_time
        if adc_raster_time is None:
            adc_raster_time = Opts.default.adc_raster_time
        if block_duration_raster is None:
            block_duration_raster = Opts.default.block_duration_raster

        if rf_dead_time is None:
            rf_dead_time = Opts.default.rf_dead_time
        if rf_raster_time is None:
            rf_raster_time = Opts.default.rf_raster_time
        if grad_raster_time is None:
            grad_raster_time = Opts.default.grad_raster_time
        if rf_ringdown_time is None:
            rf_ringdown_time = Opts.default.rf_ringdown_time
        if adc_samples_limit is None:
            adc_samples_limit = Opts.default.adc_samples_limit
        if adc_samples_divisor is None:
            adc_samples_divisor = Opts.default.adc_samples_divisor
        if B0 is None:
            B0 = Opts.default.B0

        self.max_grad = max_grad
        self.max_slew = max_slew
        self.rise_time = rise_time
        self.rf_dead_time = rf_dead_time
        self.rf_ringdown_time = rf_ringdown_time
        self.adc_dead_time = adc_dead_time
        self.adc_raster_time = adc_raster_time
        self.rf_raster_time = rf_raster_time
        self.grad_raster_time = grad_raster_time
        self.block_duration_raster = block_duration_raster
        self.adc_samples_limit = adc_samples_limit
        self.adc_samples_divisor = adc_samples_divisor
        self.gamma = gamma
        self.B0 = B0

    def set_as_default(self):
        Opts.default = self

    @classmethod
    def reset_default(cls):
        cls.default = Opts(
            max_grad=convert(from_value=40, from_unit='mT/m'),
            max_slew=convert(from_value=170, from_unit='T/m/s'),
            rf_dead_time=0,
            rf_ringdown_time=0,
            adc_dead_time=0,
            adc_raster_time=100e-9,
            rf_raster_time=1e-6,
            grad_raster_time=10e-6,
            block_duration_raster=10e-6,
            adc_samples_limit=0,
            adc_samples_divisor=4,
            gamma=42576000,
            B0=1.5,
        )

    def __str__(self) -> str:
        """
        Print a string representation of the system limits objects.
        """
        variables = vars(self)
        s = [f'{key}: {value}' for key, value in variables.items()]
        s = '\n'.join(s)
        s = 'System limits:\n' + s
        return s


Opts.reset_default()



from types import SimpleNamespace

import numpy as np


def make_delay(d: float) -> SimpleNamespace:
    """
    Creates a delay event.

    Parameters
    ----------
    d : float
        Delay time in seconds (s).

    Returns
    -------
    delay : SimpleNamespace
        Delay event.

    Raises
    ------
    ValueError
        If delay is invalid (not finite or < 0).
    """
    delay = SimpleNamespace()
    if not np.isfinite(d) or d < 0:
        raise ValueError('Delay {:.2f} ms is invalid'.format(d * 1e3))
    delay.type = 'delay'
    delay.delay = d
    return delay



from types import SimpleNamespace
from typing import Union

import numpy as np

from pypulseq import eps
from pypulseq.make_arbitrary_grad import make_arbitrary_grad
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform
from pypulseq.utils.tracing import trace, trace_enabled


def make_extended_trapezoid(
    channel: str,
    amplitudes: Union[np.ndarray, None] = None,
    convert_to_arbitrary: bool = False,
    max_grad: float = 0.0,
    max_slew: float = 0.0,
    skip_check: bool = False,
    system: Union[Opts, None] = None,
    times: Union[np.ndarray, None] = None,
) -> SimpleNamespace:
    """
    Create a gradient by specifying a set of points (amplitudes) at specified time points(times) at a given channel
    with given system limits. Returns an arbitrary gradient object.

    See Also
    --------
    - `pypulseq.Sequence.sequence.Sequence.add_block()`
    - `pypulseq.opts.Opts`
    - `pypulseq.make_trapezoid.make_trapezoid()`

    Parameters
    ----------
    channel : str
        Orientation of extended trapezoidal gradient event. Must be one of 'x', 'y' or 'z'.
    convert_to_arbitrary : bool, default=False
        Boolean flag to indicate if the extended trapezoid gradient has to be converted into an arbitrary gradient.
    amplitudes : numpy.ndarray, default=09
        Values defined at `times` time indices.
    max_grad : float, default=0
        Maximum gradient strength.
    max_slew : float, default=0
        Maximum slew rate.
    system : Opts, default=Opts()
        System limits.
    skip_check : bool, default=False
        Boolean flag to indicate if amplitude check is to be skipped.
    times : numpy.ndarray, default=np.zeros(1)
        Time points at which `amplitudes` defines amplitude values.

    Returns
    -------
    grad : SimpleNamespace
        Extended trapezoid gradient event.

    Raises
    ------
    ValueError
        If invalid `channel` is passed. Must be one of 'x', 'y' or 'z'.
        If all elements in `times` are zero.
        If elements in `times` are not in ascending order or not distinct.
        If all elements in `amplitudes` are zero.
        If first amplitude of a gradient is non-ero and does not connect to a previous block.
    """
    if amplitudes is None:
        amplitudes = np.zeros(1)

    if times is None:
        times = np.zeros(1)

    if system is None:
        system = Opts.default

    if channel not in ['x', 'y', 'z']:
        raise ValueError(f"Invalid channel. Must be one of 'x', 'y' or 'z'. Passed: {channel}")

    times = np.asarray(times)
    amplitudes = np.asarray(amplitudes)

    if len(times) != len(amplitudes):
        raise ValueError('Times and amplitudes must have the same length.')

    if np.all(times == 0):
        raise ValueError('At least one of the given times must be non-zero')

    if np.any(np.diff(times) <= 0):
        raise ValueError('Times must be in ascending order and all times must be distinct')

    if abs(round(times[-1] / system.grad_raster_time) * system.grad_raster_time - times[-1]) > eps:
        raise ValueError('The last time point must be on a gradient raster')

    if skip_check is False and times[0] > 0 and amplitudes[0] != 0:
        raise ValueError('If first amplitude of a gradient is non-zero, it must connect to previous block')

    if max_grad <= 0:
        max_grad = system.max_grad

    if max_slew <= 0:
        max_slew = system.max_slew

    if convert_to_arbitrary:
        # Represent the extended trapezoid on the regularly sampled time grid
        waveform = points_to_waveform(times=times, amplitudes=amplitudes, grad_raster_time=system.grad_raster_time)
        grad = make_arbitrary_grad(
            channel=channel,
            waveform=waveform,
            system=system,
            max_slew=max_slew,
            max_grad=max_grad,
            delay=times[0],
        )
    else:
        #  Keep the original possibly irregular sampling
        if np.any(np.abs(np.round(times / system.grad_raster_time) * system.grad_raster_time - times) > eps):
            raise ValueError(
                'All time points must be on a gradient raster or "convert_to_arbitrary" option must be used.'
            )
        grad = SimpleNamespace()
        grad.type = 'grad'
        grad.channel = channel
        grad.waveform = amplitudes
        grad.delay = round(times[0] / system.grad_raster_time) * system.grad_raster_time
        grad.tt = times - grad.delay
        grad.shape_dur = grad.tt[-1]
        grad.area = 0.5 * ((grad.tt[1:] - grad.tt[:-1]) * (grad.waveform[1:] + grad.waveform[:-1])).sum()

    grad.first = amplitudes[0]
    grad.last = amplitudes[-1]

    slew = np.diff(grad.waveform) / np.diff(grad.tt)

    if max(abs(slew)) > max_slew * (1 + eps):
        raise ValueError(f'Slew rate violation {max(abs(slew)) / max_slew * 100:.2f}%')
    if max(abs(grad.waveform)) > max_grad + eps:
        raise ValueError(f'Gradient amplitude violation {max(abs(grad.waveform)) / max_grad * 100:.2f}%')

    if trace_enabled():
        grad.trace = trace()

    return grad



import math
from copy import copy
from types import SimpleNamespace
from typing import Tuple, Union
from warnings import warn

import numpy as np

from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.make_trapezoid import make_trapezoid
from pypulseq.opts import Opts
from pypulseq.supported_labels_rf_use import get_supported_rf_uses
from pypulseq.utils.tracing import trace, trace_enabled


def make_arbitrary_rf(
    signal: np.ndarray,
    flip_angle: float,
    bandwidth: float = 0.0,
    delay: float = 0.0,
    dwell: float = 0.0,
    freq_offset: float = 0.0,
    no_signal_scaling: bool = False,
    max_grad: float = 0.0,
    max_slew: float = 0.0,
    phase_offset: float = 0.0,
    return_gz: bool = False,
    slice_thickness: float = 0.0,
    system: Union[Opts, None] = None,
    time_bw_product: float = 0.0,
    use: str = 'undefined',
    freq_ppm: float = 0.0,
    phase_ppm: float = 0.0,
    center: Union[float, None] = None,
) -> Union[SimpleNamespace, Tuple[SimpleNamespace, SimpleNamespace]]:
    """
    Create an RF pulse with the given pulse shape.

    Parameters
    ----------
    signal : numpy.ndarray
        Arbitrary waveform.
    flip_angle : float
        Flip angle in radians.
    bandwidth : float, default=0
        Bandwidth in Hertz (Hz).
    delay : float, default=0
        Delay in seconds (s) of accompanying slice select trapezoidal event.
    dwell : float, default=0
        Temporal sampling step of waveform. If set to 0, will use `system.rf_raster_time`.
    freq_offset : float, default=0
        Frequency offset in Hertz (Hz).
    no_signal_scaling : bool, default=False
        If set to True no rescaling of the RF amplitude will happen. E.g. for adiabatic pulses.
    max_grad : float, default=system.max_grad
        Maximum gradient strength of accompanying slice select trapezoidal event.
    max_slew : float, default=system.max_slew
        Maximum slew rate of accompanying slice select trapezoidal event.
    phase_offset : float, default=0
        Phase offset in Hertz (Hz).a
    return_gz : bool, default=False
        Boolean flag to indicate if slice-selective gradient has to be returned.
    slice_thickness : float, default=0
        Slice thickness (m) of accompanying slice select trapezoidal event. The slice thickness determines the area of the
        slice select event.
    system : Opts, default=Opts()
        System limits.
    time_bw_product : float, default=0
        Time-bandwidth product.
    use : str, default='undefined'
        Use of arbitrary radio-frequency pulse event.
        Must be one of 'excitation', 'refocusing', 'inversion',
        'saturation', 'preparation', 'other', 'undefined'.
    freq_ppm : float, default=0
        PPM frequency offset.
    phase_ppm : float, default=0
        PPM phase offset.
    center : float, default=None
        RF center (s).

    Returns
    -------
    rf : SimpleNamespace
        Radio-frequency pulse event with arbitrary pulse shape.
    gz : SimpleNamespace, optional
        Slice select trapezoidal gradient event accompanying the arbitrary radio-frequency pulse event.

    Raises
    ------
    ValueError
        If invalid `use` parameter is passed.
        If `signal` with ndim > 1 is passed.
        If `return_gz=True`, and `slice_thickness` and `bandwidth` are not passed.
    """
    if system is None:
        system = Opts.default

    valid_pulse_uses = get_supported_rf_uses()
    if use != '' and use not in valid_pulse_uses:
        raise ValueError(f'Invalid use parameter. Must be one of {valid_pulse_uses}. Passed: {use}')

    if dwell == 0:
        dwell = system.rf_raster_time

    signal = np.squeeze(signal)
    if signal.ndim > 1:
        raise ValueError(f'signal should have ndim=1. Passed ndim={signal.ndim}')

    if not no_signal_scaling:
        signal = signal / np.abs(np.sum(signal * dwell)) * flip_angle / (2 * np.pi)

    n_samples = len(signal)
    duration = n_samples * dwell
    t = (np.arange(1, n_samples + 1) - 0.5) * dwell

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.shape_dur = duration
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.freq_ppm = freq_ppm
    rf.phase_ppm = phase_ppm
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay
    rf.use = use

    if rf.dead_time > rf.delay:
        warn(
            f'Specified RF delay {rf.delay * 1e6:.2f} us is less than the dead time {rf.dead_time * 1e6:.0f} us. Delay was increased to the dead time.',
            stacklevel=2,
        )
        rf.delay = rf.dead_time

    if center is not None:
        rf.center = center
        if rf.center < 0:
            rf.center = 0
        if rf.center > rf.shape_dur:
            rf.center = rf.shape_dur
    else:
        rf.center, _ = calc_rf_center(rf)

    if return_gz:
        if slice_thickness <= 0:
            raise ValueError('Slice thickness must be provided.')
        if bandwidth <= 0:
            raise ValueError('Bandwidth of pulse must be provided.')

        if max_grad > 0:
            system = copy(system)
            system.max_grad = max_grad
        if max_slew > 0:
            system = copy(system)
            system.max_slew = max_slew

        if time_bw_product > 0:
            bandwidth = time_bw_product / duration

        amplitude = bandwidth / slice_thickness
        area = amplitude * duration
        gz = make_trapezoid(channel='z', system=system, flat_time=duration, flat_area=area)

        if rf.delay > gz.rise_time:
            # Round-up to gradient raster
            gz.delay = math.ceil((rf.delay - gz.rise_time) / system.grad_raster_time) * system.grad_raster_time

        if rf.delay < (gz.rise_time + gz.delay):
            rf.delay = gz.rise_time + gz.delay

    if trace_enabled():
        rf.trace = trace()

    if return_gz:
        return rf, gz
    else:
        return rf



import numpy as np
import pypulseq as pp
import pytest
from pypulseq import scale_grad
from pypulseq.opts import Opts

# Updated gradients with realistic hardware limits
grad_list = [
    pp.make_trapezoid(channel='x', amplitude=10, duration=13, max_grad=30, max_slew=200),
    pp.make_trapezoid(channel='y', amplitude=10, duration=13, max_grad=30, max_slew=200),
    pp.make_trapezoid(channel='z', amplitude=10, duration=13, max_grad=30, max_slew=200),
    pp.make_trapezoid(channel='x', amplitude=20, duration=5, max_grad=25, max_slew=150),
    pp.make_trapezoid(channel='y', amplitude=20, duration=5, max_grad=25, max_slew=150),
    pp.make_trapezoid(channel='z', amplitude=20, duration=5, max_grad=25, max_slew=150),
    pp.make_extended_trapezoid(
        'x', [0, 15, 5, 10], convert_to_arbitrary=True, times=[1, 3, 4, 7], max_grad=40, max_slew=300
    ),
    pp.make_extended_trapezoid(
        'y', [0, 15, 5, 10], convert_to_arbitrary=True, times=[1, 3, 4, 7], max_grad=40, max_slew=300
    ),
    pp.make_extended_trapezoid(
        'z', [0, 15, 5, 10], convert_to_arbitrary=True, times=[1, 3, 4, 7], max_grad=40, max_slew=300
    ),
    pp.make_extended_trapezoid(
        'x', [0, 20, 10, 15], convert_to_arbitrary=False, times=[1, 3, 4, 7], max_grad=25, max_slew=150
    ),
    pp.make_extended_trapezoid(
        'y', [0, 20, 10, 15], convert_to_arbitrary=False, times=[1, 3, 4, 7], max_grad=25, max_slew=150
    ),
    pp.make_extended_trapezoid(
        'z', [0, 20, 10, 15], convert_to_arbitrary=False, times=[1, 3, 4, 7], max_grad=25, max_slew=150
    ),
    pp.make_extended_trapezoid(
        'x', [0, 10, 5, 10], convert_to_arbitrary=False, times=[1, 2, 3, 4], max_grad=15, max_slew=80
    ),
    pp.make_extended_trapezoid(
        'y', [0, 10, 5, 10], convert_to_arbitrary=False, times=[1, 2, 3, 4], max_grad=15, max_slew=80
    ),
    pp.make_extended_trapezoid(
        'z', [0, 10, 5, 10], convert_to_arbitrary=False, times=[1, 2, 3, 4], max_grad=15, max_slew=80
    ),
]

# ----------- TEST SCALING IS CORRECT -----------


@pytest.mark.parametrize('grad', grad_list)
def test_scale_grad_correct_scaling(grad):
    scale = 0.5  # Safe scale
    system = Opts(max_grad=40, max_slew=300)
    scaled = scale_grad(grad, scale, system)

    if grad.type == 'trap':
        assert np.isclose(scaled.amplitude, grad.amplitude * scale)
        assert np.isclose(scaled.flat_area, grad.flat_area * scale)
    else:
        assert np.allclose(scaled.waveform, grad.waveform * scale)
        assert np.isclose(scaled.first, grad.first * scale)
        assert np.isclose(scaled.last, grad.last * scale)

    assert np.isclose(scaled.area, grad.area * scale)

    # ID must be removed
    assert not hasattr(scaled, 'id')


# ----------- TEST AMPLITUDE VIOLATION -----------


def test_scale_grad_violates_amplitude():
    scale = 100.0
    system = Opts(max_grad=40, max_slew=999999999)  # make sure we have infinite slew rate

    expected_failures = 0
    actual_failures = 0

    for grad in grad_list:
        if grad.type == 'trap':
            should_fail = abs(grad.amplitude) * scale > system.max_grad
        else:
            should_fail = max(abs(grad.waveform)) * scale > system.max_grad

        if should_fail:
            expected_failures += 1
            with pytest.raises(ValueError, match='maximum amplitude exceeded'):
                scale_grad(grad, scale, system)
            actual_failures += 1
        else:
            scale_grad(grad, scale, system)

    assert expected_failures == actual_failures, (
        f'Expected {expected_failures} amplitude violations, but got {actual_failures}.'
    )


# ----------- TEST SLEW RATE VIOLATION -----------


def test_scale_grad_violates_slew():
    scale = 100.0
    system = Opts(max_grad=999999999, max_slew=300)  # make sure we have infinite gradient amp

    expected_failures = 0
    actual_failures = 0

    for grad in grad_list:
        if grad.type == 'trap':
            if abs(grad.amplitude) > 1e-6:
                approx_slew = abs(grad.amplitude * scale) / min(grad.rise_time, grad.fall_time)
                should_fail = approx_slew > system.max_slew
            else:
                should_fail = False
        else:
            if max(abs(grad.waveform)) > 1e-6:
                waveform = np.array(grad.waveform) * scale
                tt = np.array(grad.tt)
                diffs = np.abs(np.diff(waveform) / np.diff(tt))
                should_fail = max(diffs) > system.max_slew
            else:
                should_fail = False

        if should_fail:
            expected_failures += 1
            with pytest.raises(ValueError, match='maximum slew rate exceeded'):
                scale_grad(grad, scale, system)
            actual_failures += 1
        else:
            scale_grad(grad, scale, system)

    assert expected_failures == actual_failures, (
        f'Expected {expected_failures} slew violations, but got {actual_failures}.'
    )



# Copyright of the Board of Trustees of Columbia University in the City of New York
from pathlib import Path
from typing import Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import scipy.io as sio
from scipy import interpolate

from pypulseq.calc_duration import calc_duration
from pypulseq.Sequence.sequence import Sequence


def _calc_SAR(Q: np.ndarray, I: np.ndarray) -> np.ndarray:
    """
    Compute the SAR output for a given Q matrix and I current values.

    Parameters
    ----------
    Q : numpy.ndarray
        Q matrix. Refer Graesslin, Ingmar, et al. "A specific absorption rate prediction concept for parallel
        transmission MR." Magnetic resonance in medicine 68.5 (2012): 1664-1674.
    I : numpy.ndarray
        I matrix, capturing the current (in Amps) on each of the transmit channels. Refer Graesslin, Ingmar, et al. "A
        specific absorption rate prediction concept for parallel transmission MR." Magnetic resonance in medicine
        68.5 (2012): 1664-1674.

    Returns
    -------
    SAR : numpy.ndarray
       Contains the SAR value for a particular Q matrix
    """
    if len(I.shape) == 1:  # Just to fit the multi-transmit case for now, TODO
        I = np.tile(I, (Q.shape[0], 1))  # Nc x Nt

    I_fact = np.divide(np.matmul(I, np.conjugate(I).T), I.shape[1])
    SAR_temp = np.multiply(Q, I_fact)
    SAR = np.abs(np.sum(SAR_temp[:]))

    return SAR


def _load_Q() -> Tuple[np.ndarray, np.ndarray]:
    """
    Load Q matrix that is precomputed based on the VHM model for 8 channels. Refer Graesslin, Ingmar, et al. "A
    specific absorption rate prediction concept for parallel transmission MR." Magnetic resonance in medicine 68.5
    (2012): 1664-1674.

    Returns
    -------
    Qtmf, Qhmf : numpy.ndarray
        Contains the Q-matrix of global SAR values for body-mass and head-mass respectively.
    """
    # Load relevant Q matrices computed from the model - this code will be integrated later - starting from E fields
    path_Q = str(Path(__file__).parent / 'QGlobal.mat')
    Q = sio.loadmat(path_Q)
    Q = Q['Q']
    val = Q[0, 0]

    Qtmf = val['Qtmf']
    Qhmf = val['Qhmf']
    return Qtmf, Qhmf


def _SAR_from_seq(seq: Sequence, Qtmf: np.ndarray, Qhmf: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute global whole body and head only SAR values for the given `seq` object.

    Parameters
    ----------
    seq : Sequence
        Sequence object to calculate for which SAR values will be calculated.
    Qtmf : numpy.ndarray
        Q-matrix of global SAR values for body-mass.
    Qhmf : numpy.ndarray
        Q-matrix of global SAR values for head-mass.

    Returns
    -------
    SAR_wbg : numpy.ndarray
        SAR values for body-mass.
    SAR_hg : numpy.ndarray
        SAR values for head-mass.
    t : numpy.ndarray
        Corresponding time points.
    """
    # Identify RF blocks and compute SAR - 10 seconds must be less than twice and 6 minutes must be less than
    # 4 (WB) and 3.2 (head-20)
    block_events = seq.block_events
    num_events = len(block_events)
    t = np.zeros(num_events)
    SAR_wbg = np.zeros(t.shape)
    SAR_hg = np.zeros(t.shape)
    t_prev = 0

    for block_counter in block_events:
        block = seq.get_block(block_counter)
        block_dur = calc_duration(block)
        t[block_counter - 1] = t_prev + block_dur
        t_prev = t[block_counter - 1]
        if hasattr(block, 'rf') and block.rf is not None:  # has rf event
            rf = block.rf
            signal = rf.signal
            # This rf could be parallel transmit as well
            SAR_wbg[block_counter] = _calc_SAR(Qtmf, signal)
            SAR_hg[block_counter] = _calc_SAR(Qhmf, signal)

    return SAR_wbg, SAR_hg, t


def _SAR_interp(SAR: np.ndarray, t: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Interpolate SAR values for one second resolution.

    Parameters
    ----------
    SAR : numpy.ndarray
        SAR values
    t : numpy.ndarray
        Current time points.

    Returns
    -------
    SAR_interp : numpy.ndarray
        Interpolated values of SAR for a temporal resolution of 1 second.
    t_sec : numpy.ndarray
        Time points at 1 second resolution.
    """
    t_sec = np.arange(1, np.floor(t[-1]) + 1, 1)
    f = interpolate.interp1d(t, SAR)
    SAR_interp = f(t_sec)
    return SAR_interp, t_sec


def _SAR_lims_check(
    SARwbg_lim_s, SARhg_lim_s, tsec
) -> Tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
]:
    """
    Check for SAR violations as compared to IEC 10 second and 6 minute averages;
    returns SAR values that are interpolated for the fixed IEC time intervals.

    Parameters
    ----------
    SARwbg_lim_s : numpy.ndarray
    SARhg_lim_s : numpy.ndarray
    tsec : numpy.ndarray

    Returns
    -------
    SAR_wbg_tensec : numpy.ndarray
    SAR_wbg_sixmin : numpy.ndarray
    SAR_hg_tensec : numpy.ndarray
    SAR_hg_sixmin : numpy.ndarray
    SAR_wbg_sixmin_peak : numpy.ndarray
    SAR_hg_sixmin_peak : numpy.ndarray
    SAR_wbg_tensec_peak : numpy.ndarray
    SAR_hg_tensec_peak : numpy.ndarray
    """
    if tsec[-1] > 10:
        six_min_threshold_wbg = 4
        ten_sec_threshold_wbg = 8

        six_min_threshold_hg = 3.2
        ten_sec_threshold_hg = 6.4

        SAR_wbg_lim_app = np.concatenate((np.zeros(5), SARwbg_lim_s, np.zeros(5)), axis=0)
        SAR_hg_lim_app = np.concatenate((np.zeros(5), SARhg_lim_s, np.zeros(5)), axis=0)

        SAR_wbg_tensec = _do_sw_sar(SAR_wbg_lim_app, tsec, 10)  # < 2  SARmax
        SAR_hg_tensec = _do_sw_sar(SAR_hg_lim_app, tsec, 10)  # < 2 SARmax
        SAR_wbg_tensec_peak = np.round(np.max(SAR_wbg_tensec), 2)
        SAR_hg_tensec_peak = np.round(np.max(SAR_hg_tensec), 2)

        if (np.max(SAR_wbg_tensec) > ten_sec_threshold_wbg) or (np.max(SAR_hg_tensec) > ten_sec_threshold_hg):
            print('Pulse exceeding 10 second Global SAR limits, increase TR')
        SAR_wbg_sixmin = 'NA'
        SAR_hg_sixmin = 'NA'
        SAR_wbg_sixmin_peak = 'NA'
        SAR_hg_sixmin_peak = 'NA'

        if tsec[-1] > 600:
            SAR_wbg_lim_app = np.concatenate((np.zeros(300), SARwbg_lim_s, np.zeros(300)), axis=0)
            SAR_hg_lim_app = np.concatenate((np.zeros(300), SARhg_lim_s, np.zeros(300)), axis=0)

            SAR_hg_sixmin = _do_sw_sar(SAR_hg_lim_app, tsec, 600)
            SAR_wbg_sixmin = _do_sw_sar(SAR_wbg_lim_app, tsec, 600)
            SAR_wbg_sixmin_peak = np.round(np.max(SAR_wbg_sixmin), 2)
            SAR_hg_sixmin_peak = np.round(np.max(SAR_hg_sixmin), 2)

            if (np.max(SAR_hg_sixmin) > six_min_threshold_wbg) or (np.max(SAR_hg_sixmin) > six_min_threshold_hg):
                print('Pulse exceeding 10 second Global SAR limits, increase TR')
    else:
        print('Need at least 10 seconds worth of sequence to calculate SAR')
        SAR_wbg_tensec = 'NA'
        SAR_wbg_sixmin = 'NA'
        SAR_hg_tensec = 'NA'
        SAR_hg_sixmin = 'NA'
        SAR_wbg_sixmin_peak = 'NA'
        SAR_hg_sixmin_peak = 'NA'
        SAR_wbg_tensec_peak = 'NA'
        SAR_hg_tensec_peak = 'NA'

    return (
        SAR_wbg_tensec,
        SAR_wbg_sixmin,
        SAR_hg_tensec,
        SAR_hg_sixmin,
        SAR_wbg_sixmin_peak,
        SAR_hg_sixmin_peak,
        SAR_wbg_tensec_peak,
        SAR_hg_tensec_peak,
    )


def _do_sw_sar(SAR: np.ndarray, tsec: np.ndarray, t: np.ndarray) -> np.ndarray:
    """
    Compute a sliding window average of SAR values.

    Parameters
    ----------
    SAR : numpy.ndarray
        SAR values.
    tsec : numpy.ndarray
        Corresponding time points at 1 second resolution.
    t : numpy.ndarray
        Corresponding time points.

    Returns
    -------
    SAR_timeavag : numpy.ndarray
        Sliding window time average of SAR values.
    """
    SAR_time_avg = np.zeros(len(tsec) + int(t))
    for instant in range(int(t / 2), int(t / 2) + (int(tsec[-1]))):  # better to go from  -sw / 2: sw / 2
        SAR_time_avg[instant] = sum(SAR[range(instant - int(t / 2), instant + int(t / 2) - 1)]) / t
    SAR_time_avg = SAR_time_avg[int(t / 2) : int(t / 2) + (int(tsec[-1]))]
    return SAR_time_avg


def calc_SAR(file: Union[str, Path, Sequence]) -> None:
    """
    Compute Global SAR values on the `.seq` object for head and whole body over the specified time averages.

    Parameters
    ----------
    file : str, Path or Sequence
        `.seq` file for which global SAR values will be computed. Can be path to `.seq` file as `str` or `Path`, or the
        `Sequence` object itself.

    Raises
    ------
    ValueError
        If `file` is a `str` or `Path` to the `.seq` file and this file does not exist on disk.
    """
    if isinstance(file, (str, Path)):
        if isinstance(file, str):
            file = Path(file)

        if file.exists() and file.is_file():
            seq_obj = Sequence()
            seq_obj.read(str(file))
            seq_obj = seq_obj
        else:
            raise ValueError('Seq file does not exist.')
    else:
        seq_obj = file

    Q_tmf, Q_hmf = _load_Q()
    SAR_wbg, SAR_hg, t = _SAR_from_seq(seq_obj, Q_tmf, Q_hmf)
    SARwbg_lim, tsec = _SAR_interp(SAR_wbg, t)
    SARhg_lim, tsec = _SAR_interp(SAR_hg, t)
    (
        SAR_wbg_tensec,
        _SAR_wbg_sixmin,
        SAR_hg_tensec,
        _SAR_hg_sixmin,
        _SAR_wbg_sixmin_peak,
        _SAR_hg_sixmin_peak,
        _SAR_wbg_tensec_peak,
        _SAR_hg_tensec_peak,
    ) = _SAR_lims_check(SARwbg_lim, SARhg_lim, tsec)

    # Plot 10 sec average SAR
    if tsec[-1] > 10:
        plt.plot(tsec, SAR_wbg_tensec, 'x-', label='Whole Body: 10sec')
        plt.plot(tsec, SAR_hg_tensec, '.-', label='Head only: 10sec')

        # plt.plot(t, SARwbg, label='Whole Body - instant')
        # plt.plot(t, SARhg, label='Whole Body - instant')

        plt.xlabel('Time (s)')
        plt.ylabel('SAR (W/kg)')
        plt.title('Global SAR  - Mass Normalized -  Whole body and head only')

        plt.legend()
        plt.grid(True)
        plt.show()



from types import SimpleNamespace

from pypulseq.block_to_events import block_to_events


def calc_duration(*args: SimpleNamespace) -> float:
    """
    Calculate the duration of an event or block.
    The duration of an event is the
    time taken by the actual event itself plus its delay.
    If multiple events are provided, the calculated duration is for a block
    comprised of these events, which is given by the maximum duration of the events.

    Although it is possible to provide events that can't be actually be combined into a block
    (e.g. multiple events of the same type), the result is still the maximum duration of all events.

    Parameters
    ----------
    args : SimpleNamespace
        Block or events.

    Returns
    -------
    duration : float
        Maximum duration of `args`.
    """
    events = block_to_events(*args)

    duration = 0
    for event in events:
        if isinstance(event, (float, int)):  # block_duration field
            assert duration <= event
            duration = event
            continue

        if not isinstance(event, (dict, SimpleNamespace)):
            raise TypeError('input(s) should be of type SimpleNamespace or a dict() in case of LABELINC or LABELSET')

        if event.type == 'delay':
            duration = max(duration, event.delay)
        elif event.type == 'rf':
            duration = max(duration, event.delay + event.shape_dur + event.ringdown_time)
        elif event.type == 'grad':
            duration = max(duration, event.delay + event.shape_dur)
        elif event.type == 'adc':
            duration = max(
                duration,
                event.delay + event.num_samples * event.dwell + event.dead_time,
            )
        elif event.type == 'trap':
            duration = max(
                duration,
                event.delay + event.rise_time + event.flat_time + event.fall_time,
            )
        elif event.type == 'output' or event.type == 'trigger':
            duration = max(duration, event.delay + event.duration)

    return duration



from types import SimpleNamespace
from typing import Union
from warnings import warn

import numpy as np

from pypulseq.opts import Opts
from pypulseq.supported_labels_rf_use import get_supported_rf_uses
from pypulseq.utils.tracing import trace, trace_enabled


def make_block_pulse(
    flip_angle: float,
    delay: float = 0.0,
    duration: Union[float, None] = None,
    bandwidth: Union[float, None] = None,
    time_bw_product: Union[float, None] = None,
    freq_offset: float = 0.0,
    phase_offset: float = 0.0,
    system: Union[Opts, None] = None,
    use: str = 'undefined',
    freq_ppm: float = 0.0,
    phase_ppm: float = 0.0,
) -> SimpleNamespace:
    """
    Create a block (RECT or hard) pulse.

    Define duration, or bandwidth, or bandwidth and time_bw_product.
    If none are provided a default 4 ms pulse will be generated.

    Parameters
    ----------
    flip_angle : float
        Flip angle in radians.
    delay : float, default=0
        Delay in seconds (s).
    duration : float, default=None
        Duration in seconds (s).
    bandwidth : float, default=None
        Bandwidth in Hertz (Hz).
        If supplied without time_bw_product duration = 1 / (4 * bandwidth)
    time_bw_product : float, default=None
        Time-bandwidth product.
        If supplied with bandwidth, duration = time_bw_product / bandwidth
    freq_offset : float, default=0
        Frequency offset in Hertz (Hz).
    phase_offset : float, default=0
        Phase offset Hertz (Hz).
    system : Opts, default=Opts()
        System limits.
    use : str, default='undefined'
        Use of radio-frequency block pulse event.
        Must be one of 'excitation', 'refocusing', 'inversion',
        'saturation', 'preparation', 'other', 'undefined'.
    freq_ppm : float, default=0
        PPM frequency offset.
    phase_ppm : float, default=0
        PPM phase offset.

    Returns
    -------
    rf : SimpleNamespace
        Radio-frequency block pulse event.

    Raises
    ------
    ValueError
        If invalid `use` parameter is passed.
        One of bandwidth or duration must be defined, but not both.
        One of bandwidth or duration must be defined and be > 0.
    """
    if system is None:
        system = Opts.default

    valid_pulse_uses = get_supported_rf_uses()
    if use != '' and use not in valid_pulse_uses:
        raise ValueError(f'Invalid use parameter. Must be one of {valid_pulse_uses}. Passed: {use}')

    if duration is None and bandwidth is None:
        warn('Using default 4 ms duration for block pulse.')
        duration = 4e-3
    elif duration is not None and bandwidth is not None and duration > 0:
        # Multiple arguments
        raise ValueError('One of bandwidth or duration must be defined, but not both.')
    elif duration is not None and duration > 0:
        # Explicitly handle this most expected case.
        # There is probably a better way of writing this if block
        pass
    elif duration is None and bandwidth is not None and bandwidth > 0:
        if time_bw_product is not None and time_bw_product > 0:
            duration = time_bw_product / bandwidth
        else:
            duration = 1 / (4 * bandwidth)
    else:
        # Invalid arguments
        raise ValueError(
            'One of bandwidth or duration must be defined and be > 0. '
            f'duration = {duration} s, bandwidth = {bandwidth} Hz.'
        )

    n_samples = round(duration / system.rf_raster_time)
    t = np.array([0, n_samples]) * system.rf_raster_time
    signal = flip_angle / (2 * np.pi) / duration * np.ones_like(t)

    rf = SimpleNamespace()
    rf.type = 'rf'
    rf.signal = signal
    rf.t = t
    rf.shape_dur = t[-1]
    rf.freq_offset = freq_offset
    rf.phase_offset = phase_offset
    rf.freq_ppm = freq_ppm
    rf.phase_ppm = phase_ppm
    rf.dead_time = system.rf_dead_time
    rf.ringdown_time = system.rf_ringdown_time
    rf.delay = delay
    rf.center = rf.shape_dur / 2
    rf.use = use

    if rf.dead_time > rf.delay:
        warn(
            f'Specified RF delay {rf.delay * 1e6:.2f} us is less than the dead time {rf.dead_time * 1e6:.0f} us. Delay was increased to the dead time.',
            stacklevel=2,
        )
        rf.delay = rf.dead_time

    if trace_enabled():
        rf.trace = trace()

    return rf



import re
from typing import Tuple


def readasc(filename: str) -> Tuple[dict, dict]:
    """
    Reads Siemens ASC ascii-formatted textfile and returns a dictionary
    structure.
    E.g. a[0].b[2][3].c = "string"
    parses into:
      asc['a'][0]['b'][2][3]['c'] = "string"

    Parameters
    ----------
    filename : str
        Filename of the ASC file.

    Returns
    -------
    asc : dict
        Dictionary of ASC part of file.
    extra : dict
        Dictionary of other fields after "ASCCONV END"
    """
    asc, extra = {}, {}

    # Read asc file and convert it into a dictionary structure
    with open(filename, 'r') as fp:
        end_of_asc = False

        for next_line in fp:
            next_line = next_line.strip()

            if next_line == '### ASCCONV END ###':  # find end of mrProt in the asc file
                end_of_asc = True

            if next_line == '' or next_line[0] == '#':
                continue

            # regex wizardry: Matches lines like 'a[0].b[2][3].c = "string" # comment'
            # Note this assumes correct formatting, e.g. does not check whether
            # brackets match.
            match = re.match(
                r'^\s*([a-zA-Z0-9\[\]\._]+)\s*\=\s*(("[^"]*"|\'[^\']\')|(\d+)|([0-9\.e\-]+))\s*((#|\/\/)(.*))?$',
                next_line,
            )

            if match:
                field_name = match[1]

                # Keep track of where to put the value: base[assign_to] = value
                if end_of_asc:
                    base = extra
                else:
                    base = asc

                assign_to = None

                # Iterate over every segment of the field name
                parts = field_name.split('.')
                for p in parts:
                    # Update base so final assignment is like: base[assign_to][p] = value
                    if assign_to is not None and assign_to not in base:
                        base[assign_to] = {}
                    if assign_to is not None:
                        base = base[assign_to]

                    # Iterate over brackets
                    start = p.find('[')
                    if start != -1:
                        name = p[:start]
                        assign_to = name

                        while start != -1:
                            stop = p.find(']', start)
                            index = int(p[start + 1 : stop])

                            # Update base so final assignment is like: base[assign_to][p][index] = value
                            if assign_to not in base:
                                base[assign_to] = {}
                            base = base[assign_to]
                            assign_to = index

                            start = p.find('[', stop)
                    else:
                        assign_to = p

                # Depending on which regex section matched we can infer the value type
                if match[3]:
                    base[assign_to] = match[3][1:-1]
                elif match[4]:
                    base[assign_to] = int(match[4])
                elif match[5]:
                    base[assign_to] = float(match[5])
                else:
                    raise RuntimeError('This should not be reached')
            elif next_line.find('=') != -1:
                raise RuntimeError(f'Bug: ASC line with an assignment was not parsed correctly: {next_line}')

    return asc, extra



import importlib.metadata
import math
import numpy as np

# =========
# VERSION
# =========
__version__ = importlib.metadata.version("pypulseq")


# =========
# BANKER'S ROUNDING FIX
# =========
def round_half_up(n, decimals=0):
    """
    Avoid banker's rounding inconsistencies; from https://realpython.com/python-rounding/#rounding-half-up
    """
    multiplier = 10**decimals
    return math.floor(abs(n) * multiplier + 0.5) / multiplier


# =========
# EPSILON (Precision of floating point numbers)
# =========

# Instead of np.finfo(np.float64).eps, which was used before, we now try to estimate our precision based on the largest
# expected value for times, amplitudes etc (we choose 1E6) and consider another factor 10 for compounding of rounding errors.
# We then round the value to the closest power of 10.
eps = 10 ** np.floor(np.log10(np.spacing(1e6) * 10))  # this is 1e-9 for np.float64


# =========
# PACKAGE-LEVEL IMPORTS
# =========
from pypulseq.SAR.SAR_calc import calc_SAR
from pypulseq.Sequence.sequence import Sequence
from pypulseq.add_gradients import add_gradients
from pypulseq.align import align
from pypulseq.calc_duration import calc_duration
from pypulseq.calc_ramp import calc_ramp
from pypulseq.calc_rf_bandwidth import calc_rf_bandwidth
from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.make_adc import make_adc, calc_adc_segments
from pypulseq.make_adiabatic_pulse import make_adiabatic_pulse
from pypulseq.make_arbitrary_grad import make_arbitrary_grad
from pypulseq.make_arbitrary_rf import make_arbitrary_rf
from pypulseq.make_block_pulse import make_block_pulse
from pypulseq.make_delay import make_delay
from pypulseq.make_soft_delay import make_soft_delay
from pypulseq.make_digital_output_pulse import make_digital_output_pulse
from pypulseq.make_extended_trapezoid import make_extended_trapezoid
from pypulseq.make_extended_trapezoid_area import make_extended_trapezoid_area
from pypulseq.make_gauss_pulse import make_gauss_pulse
from pypulseq.make_label import make_label
from pypulseq.make_sinc_pulse import make_sinc_pulse
from pypulseq.make_trapezoid import make_trapezoid
from pypulseq.sigpy_pulse_opts import SigpyPulseOpts
from pypulseq.make_trigger import make_trigger
from pypulseq.opts import Opts
from pypulseq.points_to_waveform import points_to_waveform
from pypulseq.rotate import rotate
from pypulseq.scale_grad import scale_grad
from pypulseq.split_gradient import split_gradient
from pypulseq.split_gradient_at import split_gradient_at
from pypulseq.supported_labels_rf_use import get_supported_labels
from pypulseq.traj_to_grad import traj_to_grad
from pypulseq.utils.tracing import enable_trace, disable_trace



from typing import Tuple, Union

import numpy as np

from pypulseq.opts import Opts


def traj_to_grad(k: np.ndarray, raster_time: Union[float, None] = None) -> Tuple[np.ndarray, np.ndarray]:
    """
    Convert k-space trajectory `k` into gradient waveform in compliance with `raster_time` gradient raster time.

    Parameters
    ----------
    k : numpy.ndarray
        K-space trajectory to be converted into gradient waveform.
    raster_time : float, default=Opts().grad_raster_time
        Gradient raster time.

    Returns
    -------
    g : numpy.ndarray
        Gradient waveform.
    sr : numpy.ndarray
        Slew rate.
    """
    if raster_time is None:
        raster_time = Opts.default.grad_raster_time

    # Compute finite difference for gradients in Hz/m
    g = (k[..., 1:] - k[..., :-1]) / raster_time
    # Compute the slew rate
    sr0 = (g[..., 1:] - g[..., :-1]) / raster_time

    # Gradient is now sampled between k-space points whilst the slew rate is between gradient points
    sr = np.zeros((*sr0.shape[:-1], sr0.shape[-1] + 1))
    sr[..., 0] = sr0[..., 0]
    sr[..., 1:-1] = 0.5 * (sr0[..., :-1] + sr0[..., 1:])
    sr[..., -1] = sr0[..., -1]

    return g, sr



import math
import warnings
from types import SimpleNamespace
from typing import Tuple, Union

import numpy as np

from pypulseq.calc_rf_center import calc_rf_center
from pypulseq.opts import Opts


def calc_rf_bandwidth(
    rf: SimpleNamespace,
    cutoff: float = 0.5,
    return_axis: bool = False,
    return_spectrum: bool = False,
    dw: float = 10,
    dt: Union[float, None] = None,
) -> Union[float, Tuple[float, np.ndarray], Tuple[float, np.ndarray, float]]:
    """
    Calculate the spectrum of the RF pulse. Returns the bandwidth of the pulse (calculated by a simple FFT, e.g.
    presuming a low-angle approximation) and optionally the spectrum and the frequency axis. The default for the
    optional parameter 'cutoff' is 0.5.

    Parameters
    ----------
    rf : SimpleNamespace
        RF pulse event.
    cutoff : float, default=0.5
    return_axis : bool, default=False
        Boolean flag to indicate if frequency axis of RF pulse will be returned.
    return_spectrum : bool, default=False
        Boolean flag to indicate if spectrum of RF pulse will be returned.
    dw : float, default=10
        Spectral resolution in (Hz).
    dt : Union[float, None], default=None
        Sampling time in (s). Defaults to Opts().rf_raster_time.

    Returns
    -------
    bw : float
        Bandwidth of the RF pulse.

    """
    if dt is None:
        dt = Opts().rf_raster_time

    time_center, _ = calc_rf_center(rf)

    if abs(rf.freq_ppm) > np.finfo(float).eps:
        warnings.warn(
            'calc_rf_bandwidth(): relying on the system properties, like B0 and gamma, '
            'stored in the global environment by calling pypulseq.Opts()'
        )
        sys = Opts()
        full_freq_offset = rf.freq_offset + rf.freq_ppm * 1e-6 * sys.gamma * sys.B0
    else:
        full_freq_offset = rf.freq_offset

    # Resample the pulse to a reasonable time array
    nn = round(1 / dw / dt)
    tt = np.arange(-math.floor(nn / 2), math.ceil(nn / 2) - 1) * dt

    rf_signal = rf.signal * np.exp(1j * rf.phase_offset + 2 * math.pi * full_freq_offset * rf.t)
    rfs = np.interp(xp=rf.t - time_center, fp=rf_signal, x=tt)
    spectrum = np.fft.fftshift(np.fft.fft(np.fft.fftshift(rfs)))
    w = np.arange(-math.floor(nn / 2), math.ceil(nn / 2) - 1) * dw

    w1 = __find_flank(w, spectrum, cutoff)
    w2 = __find_flank(w[::-1], spectrum[::-1], cutoff)

    bw = w2 - w1

    if return_spectrum and not return_axis:
        return bw, spectrum
    elif return_axis and not return_spectrum:
        return bw, w
    elif return_spectrum and return_axis:
        return bw, spectrum, w

    return bw


def __find_flank(x, f, c):
    m = np.max(np.abs(f))
    f = np.abs(f) / m
    i = np.argwhere(f > c)[0]

    return x[i]



import numpy as np

from pypulseq import eps
from pypulseq.convert import convert


def ext_test_report(self) -> str:
    """
    Analyze the sequence and return a text report.

    Returns
    -------
    report : str

    """
    # Find RF pulses and list flip angles
    flip_angles_deg = []
    for k in self.rf_library.data:
        lib_data = self.rf_library.data[k]
        if len(self.rf_library.type) >= k:
            rf = self.rf_from_lib_data(lib_data, self.rf_library.type[k])
        else:
            rf = self.rf_from_lib_data(lib_data)
        flip_angles_deg.append(np.abs(np.sum(rf.signal[:-1] * (rf.t[1:] - rf.t[:-1]))) * 360)

    flip_angles_deg = np.unique(flip_angles_deg)

    # Calculate TE, TR
    duration, num_blocks, event_count = self.duration()

    k_traj_adc, _, t_excitation, _, t_adc = self.calculate_kspace()
    t_excitation = np.asarray(t_excitation)

    # remove all ADC events that come before the first RF event (noise scans or alike)
    t_adc = t_adc[t_adc > t_excitation[0]]

    k_abs_adc = np.sqrt(np.sum(np.square(k_traj_adc), axis=0))
    k_abs_echo, index_echo = np.min(k_abs_adc), np.argmin(k_abs_adc)
    t_echo = t_adc[index_echo]
    if k_abs_echo > eps:
        i2check = []
        # Check if ADC k-space trajectory has elements left and right to index_echo
        if index_echo > 1:
            i2check.append(index_echo - 1)
        if index_echo < len(k_abs_adc):
            i2check.append(index_echo + 1)

        for a in range(len(i2check)):
            v_i_to_0 = -k_traj_adc[:, index_echo]
            v_i_to_t = k_traj_adc[:, i2check[a]] - k_traj_adc[:, index_echo]
            # Project v_i_to_0 to v_i_to_t
            p_vit = np.matmul(v_i_to_0, v_i_to_t) / np.square(np.linalg.norm(v_i_to_t))
            if p_vit > 0:
                # We have found a bracket for the echo and the proportionality coefficient is p_vit
                t_echo = t_adc[index_echo] * (1 - p_vit) + t_adc[i2check[a]] * p_vit

    if len(t_excitation) != 0:
        t_ex_tmp = t_excitation[t_excitation < t_echo]
        TE = t_echo - t_ex_tmp[-1]
    else:
        TE = np.nan

    if len(t_excitation) < 2:
        TR = duration
    else:
        t_ex_tmp1 = t_excitation[t_excitation > t_echo]
        if len(t_ex_tmp1) == 0:
            TR = t_ex_tmp[-1] - t_ex_tmp[-2]
        else:
            TR = t_ex_tmp1[0] - t_ex_tmp[-1]

    # Check sequence dimensionality and spatial resolution
    k_extent = np.max(np.abs(k_traj_adc), axis=1)
    k_scale = np.max(k_extent)
    is_cartesian = False
    if k_scale != 0:
        k_bins = 4e6
        k_threshold = k_scale / k_bins

        # Detect unused dimensions and delete them
        if np.any(k_extent < k_threshold):
            k_traj_adc = np.delete(k_traj_adc, np.where(k_extent < k_threshold), axis=0)
            k_extent = np.delete(k_extent, np.where(k_extent < k_threshold), axis=0)

        # Bin the k-space trajectory to detect repetitions / slices
        k_len = k_traj_adc.shape[1]
        k_repeat = np.zeros(k_len)
        k_storage = np.zeros(k_len)
        k_storage_next = 0
        k_map = {}
        keys = np.round(k_traj_adc / k_threshold).astype(np.int32)
        for i in range(k_len):
            key_string = tuple(keys[:, i])
            k_storage_ind = k_map.get(key_string)
            if k_storage_ind is None:
                k_storage_ind = k_storage_next
                k_map[key_string] = k_storage_ind
                k_storage_next += 1
            k_storage[k_storage_ind] = k_storage[k_storage_ind] + 1
            k_repeat[i] = k_storage[k_storage_ind]

        repeats_max = np.max(k_storage[:k_storage_next])
        repeats_min = np.min(k_storage[:k_storage_next])
        repeats_median = np.median(k_storage[:k_storage_next])
        repeats_unique = np.unique(k_storage[:k_storage_next])
        counts_unique = np.zeros_like(repeats_unique)
        for i in range(len(repeats_unique)):
            counts_unique[i] = np.sum(repeats_unique[i] == k_storage[: k_storage_next - 1])

        k_traj_rep1 = k_traj_adc[:, k_repeat == 1]

        k_counters = np.zeros_like(k_traj_rep1)
        dims = k_traj_rep1.shape[0]

        keys = keys[:, k_repeat == 1]
        for j in range(dims):
            k_map = {}
            k_storage = np.zeros(k_len)
            k_storage_next = 0

            for i in range(k_traj_rep1.shape[1]):
                key = keys[j, i]
                k_storage_ind = k_map.get(key)
                if k_storage_ind is None:
                    k_storage_ind = k_map.get(key + 1)
                if k_storage_ind is None:
                    k_storage_ind = k_map.get(key - 1)
                if k_storage_ind is None:
                    k_storage_ind = k_storage_next
                    k_map[key] = k_storage_ind
                    k_storage_next += 1
                    k_storage[k_storage_ind] = k_traj_rep1[j, i]
                k_counters[j, i] = k_storage_ind

        unique_k_positions = np.max(k_counters, axis=1) + 1
        is_cartesian = np.prod(unique_k_positions) == k_traj_rep1.shape[1]
    else:
        unique_k_positions = np.ones(1)

    gw_data = self.waveforms()
    gws = [np.zeros_like(x) for x in gw_data]
    ga = np.zeros(len(gw_data))
    gs = np.zeros(len(gw_data))

    common_time = np.unique(np.concatenate(gw_data, axis=1)[0])

    # catch case where no gradients are present (e.g. FID)
    if all(x.size == 0 for x in gw_data):
        gw_ct = np.zeros(0)
        gs_ct = np.zeros(0)
    else:
        gw_ct = np.zeros((len(gw_data), len(common_time)))
        gs_ct = np.zeros((len(gw_data), len(common_time) - 1))

    for gc in range(len(gw_data)):
        if gw_data[gc].shape[1] > 0:
            # Slew
            gws[gc] = (gw_data[gc][1, 1:] - gw_data[gc][1, :-1]) / (gw_data[gc][0, 1:] - gw_data[gc][0, :-1])

            # Interpolate to common time
            gw_ct[gc] = np.interp(
                x=common_time,
                xp=gw_data[gc][0, :],
                fp=gw_data[gc][1, :],
                left=0,
                right=0,
            )

            # Sometimes there are very small steps in common_time:
            #   add 1e-10 to resolve instability (adding eps is too small)
            gs_ct[gc] = (gw_ct[gc][1:] - gw_ct[gc][:-1]) / (common_time[1:] - common_time[:-1] + 1e-10)

            # Max grad/slew per channel
            ga[gc] = np.max(np.abs(gw_data[gc][1:]))
            gs[gc] = np.max(np.abs(gws[gc]))

    ga_abs = np.max(np.sqrt(np.sum(np.square(gw_ct), axis=0)))
    gs_abs = np.max(np.sqrt(np.sum(np.square(gs_ct), axis=0)))

    timing_ok, timing_error_report = self.check_timing()

    report = (
        f'Number of blocks: {num_blocks}\n'
        f'Number of events:\n'
        f'RF: {event_count[1]:6.0f}\n'
        f'Gx: {event_count[2]:6.0f}\n'
        f'Gy: {event_count[3]:6.0f}\n'
        f'Gz: {event_count[4]:6.0f}\n'
        f'ADC: {event_count[5]:6.0f}\n'
        f'Delay: {event_count[0]:6.0f}\n'
        f'Sequence duration: {duration:.6f} s\n'
        f'TE: {TE:.6f} s\n'
        f'TR: {TR:.6f} s\n'
    )
    report += 'Flip angle: ' + ('{:.02f} ' * len(flip_angles_deg)).format(*flip_angles_deg) + 'deg\n'
    report += (
        'Unique k-space positions (aka cols, rows, etc.): '
        + ('{:.0f} ' * len(unique_k_positions)).format(*unique_k_positions)
        + '\n'
    )

    if np.any(unique_k_positions > 1):
        report += f'Dimensions: {len(k_extent)}\n'
        report += ('Spatial resolution: {:.02f} mm\n' * len(k_extent)).format(*(0.5 / k_extent * 1e3))
        report += f'Repetitions/slices/contrasts: {repeats_median}; range: [{repeats_min, repeats_max}]\n'

        if is_cartesian:
            report += 'Cartesian encoding trajectory detected\n'
        else:
            report += 'Non-cartesian/irregular encoding trajectory detected (eg: EPI, spiral, radial, etc.)\n'

    ga_converted = convert(from_value=ga, from_unit='Hz/m', to_unit='mT/m', gamma=self.system.gamma)
    gs_converted = convert(from_value=gs, from_unit='Hz/m/s', to_unit='T/m/s', gamma=self.system.gamma)
    report += (
        'Max gradient: '
        + ('{:.0f} ' * len(ga)).format(*ga)
        + 'Hz/m == '
        + ('{:.02f} ' * len(ga_converted)).format(*ga_converted)
        + 'mT/m\n'
    )
    report += (
        'Max slew rate: '
        + ('{:.0f} ' * len(gs)).format(*gs)
        + 'Hz/m/s == '
        + ('{:.02f} ' * len(ga_converted)).format(*gs_converted)
        + 'T/m/s\n'
    )

    ga_abs_converted = convert(from_value=ga_abs, from_unit='Hz/m', to_unit='mT/m', gamma=self.system.gamma)
    gs_abs_converted = convert(from_value=gs_abs, from_unit='Hz/m/s', to_unit='T/m/s', gamma=self.system.gamma)
    report += f'Max absolute gradient: {ga_abs:.0f} Hz/m == {ga_abs_converted:.2f} mT/m\n'
    report += f'Max absolute slew rate: {gs_abs:g} Hz/m/s == {gs_abs_converted:.2f} T/m/s'

    if timing_ok:
        report += '\nEvent timing check passed successfully\n'
    else:
        report += f'\nEvent timing check failed with {len(timing_error_report)} errors in total. \n'
        report += 'Details of the first up to 20 timing errors:'
        max_errors = min(20, len(timing_error_report))
        for line in timing_error_report[:max_errors]:
            report += f'\n{line}'
        if len(timing_error_report) > max_errors:
            report += '\n...'

    return report
